from typing import Optional, Union

import weaviate  # type: ignore
from langchain.embeddings.base import Embeddings
from langchain.schema import BaseRetriever, Document
from langchain_community.vectorstores import VectorStore, Weaviate

from langflow import CustomComponent


class WeaviateVectorStore(CustomComponent):
    display_name: str = "Weaviate"
    description: str = "Implementation of Vector Store using Weaviate"
    documentation = "https://python.langchain.com/docs/integrations/vectorstores/weaviate"
    beta = True
    field_config = {
        "url": {"display_name": "Weaviate URL", "value": "http://localhost:8080"},
        "api_key": {
            "display_name": "API Key",
            "password": True,
            "required": False,
        },
        "index_name": {
            "display_name": "Index name",
            "required": False,
        },
        "text_key": {"display_name": "Text Key", "required": False, "advanced": True, "value": "text"},
        "documents": {"display_name": "Documents", "is_list": True},
        "embedding": {"display_name": "Embedding"},
        "attributes": {
            "display_name": "Attributes",
            "required": False,
            "is_list": True,
            "field_type": "str",
            "advanced": True,
        },
        "search_by_text": {"display_name": "Search By Text", "field_type": "bool", "advanced": True},
        "code": {"show": False},
    }

    def build(
        self,
        url: str,
        search_by_text: bool = False,
        api_key: Optional[str] = None,
        index_name: Optional[str] = None,
        text_key: str = "text",
        embedding: Optional[Embeddings] = None,
        documents: Optional[Document] = None,
        attributes: Optional[list] = None,
    ) -> Union[VectorStore, BaseRetriever]:
        if api_key:
            auth_config = weaviate.AuthApiKey(api_key=api_key)
            client = weaviate.Client(url=url, auth_client_secret=auth_config)
        else:
            client = weaviate.Client(url=url)

        def _to_pascal_case(word: str):
            if word and not word[0].isupper():
                word = word.capitalize()

            if word.isidentifier():
                return word

            word = word.replace("-", " ").replace("_", " ")
            parts = word.split()
            pascal_case_word = "".join([part.capitalize() for part in parts])

            return pascal_case_word

        index_name = _to_pascal_case(index_name) if index_name else None

        if documents is not None and embedding is not None:
            return Weaviate.from_documents(
                client=client,
                index_name=index_name,
                documents=documents,
                embedding=embedding,
                by_text=search_by_text,
            )

        return Weaviate(
            client=client,
            index_name=index_name,
            text_key=text_key,
            embedding=embedding,
            by_text=search_by_text,
            attributes=attributes if attributes is not None else [],
        )



import platform
import socket
import sys
from pathlib import Path
from typing import Optional

import typer
from dotenv import load_dotenv
from multiprocess import cpu_count  # type: ignore
from rich import box
from rich import print as rprint
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from sqlmodel import select

from langflow.main import setup_app
from langflow.services.database.utils import session_getter
from langflow.services.deps import get_db_service, get_settings_service
from langflow.services.utils import initialize_services, initialize_settings_service
from langflow.utils.logger import configure, logger

console = Console()

app = typer.Typer(no_args_is_help=True)


def get_number_of_workers(workers=None):
    if workers == -1 or workers is None:
        workers = (cpu_count() * 2) + 1
    logger.debug(f"Number of workers: {workers}")
    return workers


def display_results(results):
    """
    Display the results of the migration.
    """
    for table_results in results:
        table = Table(title=f"Migration {table_results.table_name}")
        table.add_column("Name")
        table.add_column("Type")
        table.add_column("Status")

        for result in table_results.results:
            status = "Success" if result.success else "Failure"
            color = "green" if result.success else "red"
            table.add_row(result.name, result.type, f"[{color}]{status}[/{color}]")

        console.print(table)
        console.print()  # Print a new line


def set_var_for_macos_issue():
    # OBJC_DISABLE_INITIALIZE_FORK_SAFETY=YES
    # we need to set this var is we are running on MacOS
    # otherwise we get an error when running gunicorn

    if platform.system() in ["Darwin"]:
        import os

        os.environ["OBJC_DISABLE_INITIALIZE_FORK_SAFETY"] = "YES"
        # https://stackoverflow.com/questions/75747888/uwsgi-segmentation-fault-with-flask-python-app-behind-nginx-after-running-for-2 # noqa
        os.environ["no_proxy"] = "*"  # to avoid error with gunicorn
        logger.debug("Set OBJC_DISABLE_INITIALIZE_FORK_SAFETY to YES to avoid error")


def update_settings(
    config: str,
    cache: Optional[str] = None,
    dev: bool = False,
    remove_api_keys: bool = False,
    components_path: Optional[Path] = None,
    store: bool = False,
):
    """Update the settings from a config file."""

    # Check for database_url in the environment variables
    initialize_settings_service()
    settings_service = get_settings_service()
    if config:
        logger.debug(f"Loading settings from {config}")
        settings_service.settings.update_from_yaml(config, dev=dev)
    if remove_api_keys:
        logger.debug(f"Setting remove_api_keys to {remove_api_keys}")
        settings_service.settings.update_settings(REMOVE_API_KEYS=remove_api_keys)
    if cache:
        logger.debug(f"Setting cache to {cache}")
        settings_service.settings.update_settings(CACHE=cache)
    if components_path:
        logger.debug(f"Adding component path {components_path}")
        settings_service.settings.update_settings(COMPONENTS_PATH=components_path)
    if not store:
        logger.debug("Setting store to False")
        settings_service.settings.update_settings(STORE=False)


def version_callback(value: bool):
    """
    Show the version and exit.
    """
    from langflow import __version__

    if value:
        typer.echo(f"Langflow Version: {__version__}")
        raise typer.Exit()


@app.callback()
def main_entry_point(
    version: bool = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Show the version and exit.",
    ),
):
    """
    Main entry point for the Langflow CLI.
    """
    pass


@app.command()
def run(
    host: str = typer.Option("127.0.0.1", help="Host to bind the server to.", envvar="LANGFLOW_HOST"),
    workers: int = typer.Option(1, help="Number of worker processes.", envvar="LANGFLOW_WORKERS"),
    timeout: int = typer.Option(300, help="Worker timeout in seconds."),
    port: int = typer.Option(7860, help="Port to listen on.", envvar="LANGFLOW_PORT"),
    components_path: Optional[Path] = typer.Option(
        Path(__file__).parent / "components",
        help="Path to the directory containing custom components.",
        envvar="LANGFLOW_COMPONENTS_PATH",
    ),
    config: str = typer.Option(Path(__file__).parent / "config.yaml", help="Path to the configuration file."),
    # .env file param
    env_file: Path = typer.Option(None, help="Path to the .env file containing environment variables."),
    log_level: str = typer.Option("critical", help="Logging level.", envvar="LANGFLOW_LOG_LEVEL"),
    log_file: Path = typer.Option("logs/langflow.log", help="Path to the log file.", envvar="LANGFLOW_LOG_FILE"),
    cache: Optional[str] = typer.Option(
        envvar="LANGFLOW_LANGCHAIN_CACHE",
        help="Type of cache to use. (InMemoryCache, SQLiteCache)",
        default=None,
    ),
    dev: bool = typer.Option(False, help="Run in development mode (may contain bugs)"),
    path: str = typer.Option(
        None,
        help="Path to the frontend directory containing build files. This is for development purposes only.",
        envvar="LANGFLOW_FRONTEND_PATH",
    ),
    open_browser: bool = typer.Option(
        True,
        help="Open the browser after starting the server.",
        envvar="LANGFLOW_OPEN_BROWSER",
    ),
    remove_api_keys: bool = typer.Option(
        False,
        help="Remove API keys from the projects saved in the database.",
        envvar="LANGFLOW_REMOVE_API_KEYS",
    ),
    backend_only: bool = typer.Option(
        False,
        help="Run only the backend server without the frontend.",
        envvar="LANGFLOW_BACKEND_ONLY",
    ),
    store: bool = typer.Option(
        True,
        help="Enables the store features.",
        envvar="LANGFLOW_STORE",
    ),
):
    """
    Run the Langflow.
    """

    set_var_for_macos_issue()
    # override env variables with .env file

    if env_file:
        load_dotenv(env_file, override=True)

    configure(log_level=log_level, log_file=log_file)
    update_settings(
        config,
        dev=dev,
        remove_api_keys=remove_api_keys,
        cache=cache,
        components_path=components_path,
        store=store,
    )
    # create path object if path is provided
    static_files_dir: Optional[Path] = Path(path) if path else None
    app = setup_app(static_files_dir=static_files_dir, backend_only=backend_only)
    # check if port is being used
    if is_port_in_use(port, host):
        port = get_free_port(port)

    options = {
        "bind": f"{host}:{port}",
        "workers": get_number_of_workers(workers),
        "timeout": timeout,
    }

    # Define an env variable to know if we are just testing the server
    if "pytest" in sys.modules:
        return

    if platform.system() in ["Windows"]:
        # Run using uvicorn on MacOS and Windows
        # Windows doesn't support gunicorn
        # MacOS requires an env variable to be set to use gunicorn
        run_on_windows(host, port, log_level, options, app)
    else:
        # Run using gunicorn on Linux
        run_on_mac_or_linux(host, port, log_level, options, app)


def run_on_mac_or_linux(host, port, log_level, options, app):
    print_banner(host, port)
    run_langflow(host, port, log_level, options, app)


def run_on_windows(host, port, log_level, options, app):
    """
    Run the Langflow server on Windows.
    """
    print_banner(host, port)
    run_langflow(host, port, log_level, options, app)


def is_port_in_use(port, host="localhost"):
    """
    Check if a port is in use.

    Args:
        port (int): The port number to check.
        host (str): The host to check the port on. Defaults to 'localhost'.

    Returns:
        bool: True if the port is in use, False otherwise.
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex((host, port)) == 0


def get_free_port(port):
    """
    Given a used port, find a free port.

    Args:
        port (int): The port number to check.

    Returns:
        int: A free port number.
    """
    while is_port_in_use(port):
        port += 1
    return port


def print_banner(host, port):
    # console = Console()

    word = "Langflow"
    colors = ["#3300cc"]

    styled_word = ""

    for i, char in enumerate(word):
        color = colors[i % len(colors)]
        styled_word += f"[{color}]{char}[/]"

    # Title with emojis and gradient text
    title = (
        f"[bold]Welcome to :chains: {styled_word} [/bold]\n\n"
        f"Access [link=http://{host}:{port}]http://{host}:{port}[/link]"
    )
    info_text = (
        "Collaborate, and contribute at our "
        "[bold][link=https://github.com/logspace-ai/langflow]GitHub Repo[/link][/bold] :rocket:"
    )

    # Create a panel with the title and the info text, and a border around it
    panel = Panel(f"{title}\n{info_text}", box=box.ROUNDED, border_style="blue", expand=False)

    # Print the banner with a separator line before and after
    rprint(panel)


def run_langflow(host, port, log_level, options, app):
    """
    Run Langflow server on localhost
    """
    try:
        if platform.system() in ["Windows", "Darwin"]:
            # Run using uvicorn on MacOS and Windows
            # Windows doesn't support gunicorn
            # MacOS requires an env variable to be set to use gunicorn

            import uvicorn

            uvicorn.run(
                app,
                host=host,
                port=port,
                log_level=log_level,
            )
        else:
            from langflow.server import LangflowApplication

            LangflowApplication(app, options).run()
    except KeyboardInterrupt:
        logger.info("Shutting down server")
        sys.exit(0)
    except Exception as e:
        logger.exception(e)
        sys.exit(1)


@app.command()
def superuser(
    username: str = typer.Option(..., prompt=True, help="Username for the superuser."),
    password: str = typer.Option(..., prompt=True, hide_input=True, help="Password for the superuser."),
    log_level: str = typer.Option("critical", help="Logging level.", envvar="LANGFLOW_LOG_LEVEL"),
):
    """
    Create a superuser.
    """
    configure(log_level=log_level)
    initialize_services()
    db_service = get_db_service()
    with session_getter(db_service) as session:
        from langflow.services.auth.utils import create_super_user

        if create_super_user(db=session, username=username, password=password):
            # Verify that the superuser was created
            from langflow.services.database.models.user.model import User

            user: User = session.exec(select(User).where(User.username == username)).first()
            if user is None or not user.is_superuser:
                typer.echo("Superuser creation failed.")
                return

            typer.echo("Superuser created successfully.")

        else:
            typer.echo("Superuser creation failed.")


@app.command()
def migration(
    test: bool = typer.Option(True, help="Run migrations in test mode."),
    fix: bool = typer.Option(
        False,
        help="Fix migrations. This is a destructive operation, and should only be used if you know what you are doing.",
    ),
):
    """
    Run or test migrations.
    """
    if fix:
        if not typer.confirm(
            "This will delete all data necessary to fix migrations. Are you sure you want to continue?"
        ):
            raise typer.Abort()

    initialize_services(fix_migration=fix)
    db_service = get_db_service()
    if not test:
        db_service.run_migrations()
    results = db_service.run_migrations_test()
    display_results(results)


def main():
    app()


if __name__ == "__main__":
    main()



from typing import Any, ClassVar, Dict, List, Optional, Type

from langchain_community import retrievers
from langflow.interface.base import LangChainTypeCreator
from langflow.interface.importing.utils import import_class
from langflow.services.deps import get_settings_service
from langflow.template.frontend_node.retrievers import RetrieverFrontendNode
from langflow.utils.util import build_template_from_class, build_template_from_method
from loguru import logger


class RetrieverCreator(LangChainTypeCreator):
    type_name: str = "retrievers"

    from_method_nodes: ClassVar[Dict] = {
        "MultiQueryRetriever": "from_llm",
        "ZepRetriever": "__init__",
    }

    @property
    def frontend_node_class(self) -> Type[RetrieverFrontendNode]:
        return RetrieverFrontendNode

    @property
    def type_to_loader_dict(self) -> Dict:
        if self.type_dict is None:
            self.type_dict: dict[str, Any] = {
                retriever_name: import_class(f"langchain_community.retrievers.{retriever_name}")
                for retriever_name in retrievers.__all__
            }
        return self.type_dict

    def get_signature(self, name: str) -> Optional[Dict]:
        """Get the signature of an embedding."""
        try:
            if name in self.from_method_nodes:
                return build_template_from_method(
                    name,
                    type_to_cls_dict=self.type_to_loader_dict,
                    method_name=self.from_method_nodes[name],
                )
            else:
                return build_template_from_class(name, type_to_cls_dict=self.type_to_loader_dict)
        except ValueError as exc:
            raise ValueError(f"Retriever {name} not found") from exc
        except AttributeError as exc:
            logger.error(f"Retriever {name} not loaded: {exc}")
            return None

    def to_list(self) -> List[str]:
        settings_service = get_settings_service()
        return [
            retriever
            for retriever in self.type_to_loader_dict.keys()
            if retriever in settings_service.settings.RETRIEVERS or settings_service.settings.DEV
        ]


retriever_creator = RetrieverCreator()



from typing import Optional

from langchain_community.chat_models.baidu_qianfan_endpoint import QianfanChatEndpoint
from langchain.llms.base import BaseLLM
from pydantic.v1 import SecretStr

from langflow import CustomComponent


class QianfanChatEndpointComponent(CustomComponent):
    display_name: str = "QianfanChatEndpoint"
    description: str = (
        "Baidu Qianfan chat models. Get more detail from "
        "https://python.langchain.com/docs/integrations/chat/baidu_qianfan_endpoint."
    )

    def build_config(self):
        return {
            "model": {
                "display_name": "Model Name",
                "options": [
                    "ERNIE-Bot",
                    "ERNIE-Bot-turbo",
                    "BLOOMZ-7B",
                    "Llama-2-7b-chat",
                    "Llama-2-13b-chat",
                    "Llama-2-70b-chat",
                    "Qianfan-BLOOMZ-7B-compressed",
                    "Qianfan-Chinese-Llama-2-7B",
                    "ChatGLM2-6B-32K",
                    "AquilaChat-7B",
                ],
                "info": "https://python.langchain.com/docs/integrations/chat/baidu_qianfan_endpoint",
                "required": True,
            },
            "qianfan_ak": {
                "display_name": "Qianfan Ak",
                "required": True,
                "password": True,
                "info": "which you could get from  https://cloud.baidu.com/product/wenxinworkshop",
            },
            "qianfan_sk": {
                "display_name": "Qianfan Sk",
                "required": True,
                "password": True,
                "info": "which you could get from  https://cloud.baidu.com/product/wenxinworkshop",
            },
            "top_p": {
                "display_name": "Top p",
                "field_type": "float",
                "info": "Model params, only supported in ERNIE-Bot and ERNIE-Bot-turbo",
                "value": 0.8,
            },
            "temperature": {
                "display_name": "Temperature",
                "field_type": "float",
                "info": "Model params, only supported in ERNIE-Bot and ERNIE-Bot-turbo",
                "value": 0.95,
            },
            "penalty_score": {
                "display_name": "Penalty Score",
                "field_type": "float",
                "info": "Model params, only supported in ERNIE-Bot and ERNIE-Bot-turbo",
                "value": 1.0,
            },
            "endpoint": {
                "display_name": "Endpoint",
                "info": "Endpoint of the Qianfan LLM, required if custom model used.",
            },
            "code": {"show": False},
        }

    def build(
        self,
        model: str = "ERNIE-Bot-turbo",
        qianfan_ak: Optional[str] = None,
        qianfan_sk: Optional[str] = None,
        top_p: Optional[float] = None,
        temperature: Optional[float] = None,
        penalty_score: Optional[float] = None,
        endpoint: Optional[str] = None,
    ) -> BaseLLM:
        try:
            output = QianfanChatEndpoint(  # type: ignore
                model=model,
                qianfan_ak=SecretStr(qianfan_ak) if qianfan_ak else None,
                qianfan_sk=SecretStr(qianfan_sk) if qianfan_sk else None,
                top_p=top_p,
                temperature=temperature,
                penalty_score=penalty_score,
                endpoint=endpoint,
            )
        except Exception as e:
            raise ValueError("Could not connect to Baidu Qianfan API.") from e
        return output  # type: ignore



from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

from langflow.services.database.models.api_key.model import ApiKeyRead
from langflow.services.database.models.base import orjson_dumps
from langflow.services.database.models.flow import FlowCreate, FlowRead
from langflow.services.database.models.user import UserRead
from pydantic import BaseModel, Field, field_validator


class BuildStatus(Enum):
    """Status of the build."""

    SUCCESS = "success"
    FAILURE = "failure"
    STARTED = "started"
    IN_PROGRESS = "in_progress"


class GraphData(BaseModel):
    """Data inside the exported flow."""

    nodes: List[Dict[str, Any]]
    edges: List[Dict[str, Any]]


class ExportedFlow(BaseModel):
    """Exported flow from Langflow."""

    description: str
    name: str
    id: str
    data: GraphData


class InputRequest(BaseModel):
    input: dict


class TweaksRequest(BaseModel):
    tweaks: Optional[Dict[str, Dict[str, str]]] = Field(default_factory=dict)


class UpdateTemplateRequest(BaseModel):
    template: dict


class TaskResponse(BaseModel):
    """Task response schema."""

    id: Optional[str] = Field(None)
    href: Optional[str] = Field(None)


class ProcessResponse(BaseModel):
    """Process response schema."""

    result: Any
    status: Optional[str] = None
    task: Optional[TaskResponse] = None
    session_id: Optional[str] = None
    backend: Optional[str] = None


class PreloadResponse(BaseModel):
    """Preload response schema."""

    session_id: Optional[str] = None
    is_clear: Optional[bool] = None


# TaskStatusResponse(
#         status=task.status, result=task.result if task.ready() else None
#     )
class TaskStatusResponse(BaseModel):
    """Task status response schema."""

    status: str
    result: Optional[Any] = None


class ChatMessage(BaseModel):
    """Chat message schema."""

    is_bot: bool = False
    message: Union[str, None, dict] = None
    chatKey: Optional[str] = None
    type: str = "human"


class ChatResponse(ChatMessage):
    """Chat response schema."""

    intermediate_steps: str

    type: str
    is_bot: bool = True
    files: list = []

    @field_validator("type")
    @classmethod
    def validate_message_type(cls, v):
        if v not in ["start", "stream", "end", "error", "info", "file"]:
            raise ValueError("type must be start, stream, end, error, info, or file")
        return v


class PromptResponse(ChatMessage):
    """Prompt response schema."""

    prompt: str
    type: str = "prompt"
    is_bot: bool = True


class FileResponse(ChatMessage):
    """File response schema."""

    data: Any = None
    data_type: str
    type: str = "file"
    is_bot: bool = True

    @field_validator("data_type")
    @classmethod
    def validate_data_type(cls, v):
        if v not in ["image", "csv"]:
            raise ValueError("data_type must be image or csv")
        return v


class FlowListCreate(BaseModel):
    flows: List[FlowCreate]


class FlowListRead(BaseModel):
    flows: List[FlowRead]


class InitResponse(BaseModel):
    flowId: str


class BuiltResponse(BaseModel):
    built: bool


class UploadFileResponse(BaseModel):
    """Upload file response schema."""

    flowId: str
    file_path: Path


class StreamData(BaseModel):
    event: str
    data: dict

    def __str__(self) -> str:
        return f"event: {self.event}\ndata: {orjson_dumps(self.data, indent_2=False)}\n\n"


class CustomComponentCode(BaseModel):
    code: str
    field: Optional[str] = None
    frontend_node: Optional[dict] = None


class CustomComponentResponseError(BaseModel):
    detail: str
    traceback: str


class ComponentListCreate(BaseModel):
    flows: List[FlowCreate]


class ComponentListRead(BaseModel):
    flows: List[FlowRead]


class UsersResponse(BaseModel):
    total_count: int
    users: List[UserRead]


class ApiKeyResponse(BaseModel):
    id: str
    api_key: str
    name: str
    created_at: str
    last_used_at: str


class ApiKeysResponse(BaseModel):
    total_count: int
    user_id: UUID
    api_keys: List[ApiKeyRead]


class CreateApiKeyRequest(BaseModel):
    name: str


class Token(BaseModel):
    access_token: str
    refresh_token: str
    token_type: str


class ApiKeyCreateRequest(BaseModel):
    api_key: str



from langchain.llms.base import BaseLLM
from langchain.prompts import PromptTemplate
from langchain_core.documents import Document

from langflow import CustomComponent


class PromptRunner(CustomComponent):
    display_name: str = "Prompt Runner"
    description: str = "Run a Chain with the given PromptTemplate"
    beta: bool = True
    field_config = {
        "llm": {"display_name": "LLM"},
        "prompt": {
            "display_name": "Prompt Template",
            "info": "Make sure the prompt has all variables filled.",
        },
        "code": {"show": False},
    }

    def build(self, llm: BaseLLM, prompt: PromptTemplate, inputs: dict = {}) -> Document:
        chain = prompt | llm
        # The input is an empty dict because the prompt is already filled
        result = chain.invoke(input=inputs)
        if hasattr(result, "content"):
            result = result.content
        self.repr_value = result
        return Document(page_content=str(result))



from typing import Optional

from langchain.llms.base import BaseLLM
from langchain.llms.bedrock import Bedrock
from langflow import CustomComponent


class AmazonBedrockComponent(CustomComponent):
    display_name: str = "Amazon Bedrock"
    description: str = "LLM model from Amazon Bedrock."

    def build_config(self):
        return {
            "model_id": {
                "display_name": "Model Id",
                "options": [
                    "ai21.j2-grande-instruct",
                    "ai21.j2-jumbo-instruct",
                    "ai21.j2-mid",
                    "ai21.j2-mid-v1",
                    "ai21.j2-ultra",
                    "ai21.j2-ultra-v1",
                    "anthropic.claude-instant-v1",
                    "anthropic.claude-v1",
                    "anthropic.claude-v2",
                    "cohere.command-text-v14",
                ],
            },
            "credentials_profile_name": {"display_name": "Credentials Profile Name"},
            "streaming": {"display_name": "Streaming", "field_type": "bool"},
            "endpoint_url": {"display_name": "Endpoint URL"},
            "region_name": {"display_name": "Region Name"},
            "model_kwargs": {"display_name": "Model Kwargs"},
            "cache": {"display_name": "Cache"},
            "code": {"advanced": True},
        }

    def build(
        self,
        model_id: str = "anthropic.claude-instant-v1",
        credentials_profile_name: Optional[str] = None,
        region_name: Optional[str] = None,
        model_kwargs: Optional[dict] = None,
        endpoint_url: Optional[str] = None,
        streaming: bool = False,
        cache: Optional[bool] = None,
    ) -> BaseLLM:
        try:
            output = Bedrock(
                credentials_profile_name=credentials_profile_name,
                model_id=model_id,
                region_name=region_name,
                model_kwargs=model_kwargs,
                endpoint_url=endpoint_url,
                streaming=streaming,
                cache=cache,
            )  # type: ignore
        except Exception as e:
            raise ValueError("Could not connect to AmazonBedrock API.") from e
        return output



from langchain_core.vectorstores import VectorStoreRetriever

from langflow import CustomComponent
from langflow.field_typing import VectorStore


class VectoStoreRetrieverComponent(CustomComponent):
    display_name = "VectorStore Retriever"
    description = "A vector store retriever"

    def build_config(self):
        return {
            "vectorstore": {"display_name": "Vector Store", "type": VectorStore},
        }

    def build(self, vectorstore: VectorStore) -> VectorStoreRetriever:
        return vectorstore.as_retriever()



from typing import List
from langflow import CustomComponent
import json
from langchain.schema import BaseRetriever
from langchain.schema.vectorstore import VectorStore
from langchain.base_language import BaseLanguageModel
from langchain.retrievers.self_query.base import SelfQueryRetriever
from langchain.chains.query_constructor.base import AttributeInfo


class VectaraSelfQueryRetriverComponent(CustomComponent):
    """
    A custom component for implementing Vectara Self Query Retriever using a vector store.
    """

    display_name: str = "Vectara Self Query Retriever for Vectara Vector Store"
    description: str = "Implementation of Vectara Self Query Retriever"
    documentation = "https://python.langchain.com/docs/integrations/retrievers/self_query/vectara_self_query"
    beta = True

    field_config = {
        "code": {"show": True},
        "vectorstore": {"display_name": "Vector Store", "info": "Input Vectara Vectore Store"},
        "llm": {"display_name": "LLM", "info": "For self query retriever"},
        "document_content_description": {
            "display_name": "Document Content Description",
            "info": "For self query retriever",
        },
        "metadata_field_info": {
            "display_name": "Metadata Field Info",
            "info": 'Each metadata field info is a string in the form of key value pair dictionary containing additional search metadata.\nExample input: {"name":"speech","description":"what name of the speech","type":"string or list[string]"}.\nThe keys should remain constant(name, description, type)',
        },
    }

    def build(
        self,
        vectorstore: VectorStore,
        document_content_description: str,
        llm: BaseLanguageModel,
        metadata_field_info: List[str],
    ) -> BaseRetriever:
        metadata_field_obj = []

        for meta in metadata_field_info:
            meta_obj = json.loads(meta)
            if "name" not in meta_obj or "description" not in meta_obj or "type" not in meta_obj:
                raise Exception("Incorrect metadata field info format.")
            attribute_info = AttributeInfo(
                name=meta_obj["name"],
                description=meta_obj["description"],
                type=meta_obj["type"],
            )
            metadata_field_obj.append(attribute_info)

        return SelfQueryRetriever.from_llm(
            llm, vectorstore, document_content_description, metadata_field_obj, verbose=True
        )



from typing import ClassVar, Dict, List, Optional, Type

from langchain import output_parsers

from langflow.interface.base import LangChainTypeCreator
from langflow.interface.importing.utils import import_class
from langflow.services.deps import get_settings_service

from langflow.template.frontend_node.output_parsers import OutputParserFrontendNode
from loguru import logger
from langflow.utils.util import build_template_from_class, build_template_from_method


class OutputParserCreator(LangChainTypeCreator):
    type_name: str = "output_parsers"
    from_method_nodes: ClassVar[Dict] = {
        "StructuredOutputParser": "from_response_schemas",
    }

    @property
    def frontend_node_class(self) -> Type[OutputParserFrontendNode]:
        return OutputParserFrontendNode

    @property
    def type_to_loader_dict(self) -> Dict:
        if self.type_dict is None:
            settings_service = get_settings_service()
            self.type_dict = {
                output_parser_name: import_class(f"langchain.output_parsers.{output_parser_name}")
                # if output_parser_name is not lower case it is a class
                for output_parser_name in output_parsers.__all__
            }
            self.type_dict = {
                name: output_parser
                for name, output_parser in self.type_dict.items()
                if name in settings_service.settings.OUTPUT_PARSERS or settings_service.settings.DEV
            }
        return self.type_dict

    def get_signature(self, name: str) -> Optional[Dict]:
        try:
            if name in self.from_method_nodes:
                return build_template_from_method(
                    name,
                    type_to_cls_dict=self.type_to_loader_dict,
                    method_name=self.from_method_nodes[name],
                )
            else:
                return build_template_from_class(
                    name,
                    type_to_cls_dict=self.type_to_loader_dict,
                )
        except ValueError as exc:
            # raise ValueError("OutputParser not found") from exc
            logger.error(f"OutputParser {name} not found: {exc}")
        except AttributeError as exc:
            logger.error(f"OutputParser {name} not loaded: {exc}")
        return None

    def to_list(self) -> List[str]:
        return list(self.type_to_loader_dict.keys())


output_parser_creator = OutputParserCreator()



import base64
import contextlib
import hashlib
import os
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict

from fastapi import UploadFile
from platformdirs import user_cache_dir

if TYPE_CHECKING:
    from langflow.api.v1.schemas import BuildStatus

CACHE: Dict[str, Any] = {}

CACHE_DIR = user_cache_dir("langflow", "langflow")

PREFIX = "langflow_cache"


def create_cache_folder(func):
    def wrapper(*args, **kwargs):
        # Get the destination folder
        cache_path = Path(CACHE_DIR) / PREFIX

        # Create the destination folder if it doesn't exist
        os.makedirs(cache_path, exist_ok=True)

        return func(*args, **kwargs)

    return wrapper


@create_cache_folder
def clear_old_cache_files(max_cache_size: int = 3):
    cache_dir = Path(tempfile.gettempdir()) / PREFIX
    cache_files = list(cache_dir.glob("*.dill"))

    if len(cache_files) > max_cache_size:
        cache_files_sorted_by_mtime = sorted(cache_files, key=lambda x: x.stat().st_mtime, reverse=True)

        for cache_file in cache_files_sorted_by_mtime[max_cache_size:]:
            with contextlib.suppress(OSError):
                os.remove(cache_file)


def filter_json(json_data):
    filtered_data = json_data.copy()

    # Remove 'viewport' and 'chatHistory' keys
    if "viewport" in filtered_data:
        del filtered_data["viewport"]
    if "chatHistory" in filtered_data:
        del filtered_data["chatHistory"]

    # Filter nodes
    if "nodes" in filtered_data:
        for node in filtered_data["nodes"]:
            if "position" in node:
                del node["position"]
            if "positionAbsolute" in node:
                del node["positionAbsolute"]
            if "selected" in node:
                del node["selected"]
            if "dragging" in node:
                del node["dragging"]

    return filtered_data


@create_cache_folder
def save_binary_file(content: str, file_name: str, accepted_types: list[str]) -> str:
    """
    Save a binary file to the specified folder.

    Args:
        content: The content of the file as a bytes object.
        file_name: The name of the file, including its extension.

    Returns:
        The path to the saved file.
    """
    if not any(file_name.endswith(suffix) for suffix in accepted_types):
        raise ValueError(f"File {file_name} is not accepted")

    # Get the destination folder
    cache_path = Path(CACHE_DIR) / PREFIX
    if not content:
        raise ValueError("Please, reload the file in the loader.")
    data = content.split(",")[1]
    decoded_bytes = base64.b64decode(data)

    # Create the full file path
    file_path = os.path.join(cache_path, file_name)

    # Save the binary content to the file
    with open(file_path, "wb") as file:
        file.write(decoded_bytes)

    return file_path


@create_cache_folder
def save_uploaded_file(file: UploadFile, folder_name):
    """
    Save an uploaded file to the specified folder with a hash of its content as the file name.

    Args:
        file: The uploaded file object.
        folder_name: The name of the folder to save the file in.

    Returns:
        The path to the saved file.
    """
    cache_path = Path(CACHE_DIR)
    folder_path = cache_path / folder_name
    filename = file.filename
    if isinstance(filename, str) or isinstance(filename, Path):
        file_extension = Path(filename).suffix
    else:
        file_extension = ""
    file_object = file.file

    # Create the folder if it doesn't exist
    if not folder_path.exists():
        folder_path.mkdir()

    # Create a hash of the file content
    sha256_hash = hashlib.sha256()
    # Reset the file cursor to the beginning of the file
    file_object.seek(0)
    # Iterate over the uploaded file in small chunks to conserve memory
    while chunk := file_object.read(8192):  # Read 8KB at a time (adjust as needed)
        sha256_hash.update(chunk)

    # Use the hex digest of the hash as the file name
    hex_dig = sha256_hash.hexdigest()
    file_name = f"{hex_dig}{file_extension}"

    # Reset the file cursor to the beginning of the file
    file_object.seek(0)

    # Save the file with the hash as its name
    file_path = folder_path / file_name
    with open(file_path, "wb") as new_file:
        while chunk := file_object.read(8192):
            new_file.write(chunk)

    return file_path


def update_build_status(cache_service, flow_id: str, status: "BuildStatus"):
    cached_flow = cache_service[flow_id]
    if cached_flow is None:
        raise ValueError(f"Flow {flow_id} not found in cache")
    cached_flow["status"] = status
    cache_service[flow_id] = cached_flow
    cached_flow["status"] = status
    cache_service[flow_id] = cached_flow



"""


Revision ID: 1ef9c4f3765d
Revises: fd531f8868b1
Create Date: 2023-12-04 15:00:27.968998

"""

from typing import Sequence, Union

import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "1ef9c4f3765d"
down_revision: Union[str, None] = "fd531f8868b1"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    try:
        with op.batch_alter_table("apikey", schema=None) as batch_op:
            batch_op.alter_column(
                "name", existing_type=sqlmodel.sql.sqltypes.AutoString(), nullable=True
            )
    except Exception as e:
        pass
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    try:
        with op.batch_alter_table("apikey", schema=None) as batch_op:
            batch_op.alter_column("name", existing_type=sa.VARCHAR(), nullable=False)
    except Exception as e:
        pass
    # ### end Alembic commands ###



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from asgiref.sync import async_to_sync
from celery.exceptions import SoftTimeLimitExceeded  # type: ignore
from langflow.core.celery_app import celery_app
from langflow.processing.process import Result, generate_result, process_inputs
from langflow.services.deps import get_session_service
from langflow.services.manager import initialize_session_service
from loguru import logger
from rich import print

if TYPE_CHECKING:
    from langflow.graph.vertex.base import Vertex


@celery_app.task(acks_late=True)
def test_celery(word: str) -> str:
    return f"test task return {word}"


@celery_app.task(bind=True, soft_time_limit=30, max_retries=3)
def build_vertex(self, vertex: "Vertex") -> "Vertex":
    """
    Build a vertex
    """
    try:
        vertex.task_id = self.request.id
        async_to_sync(vertex.build)()
        return vertex
    except SoftTimeLimitExceeded as e:
        raise self.retry(exc=SoftTimeLimitExceeded("Task took too long"), countdown=2) from e


@celery_app.task(acks_late=True)
def process_graph_cached_task(
    data_graph: Dict[str, Any],
    inputs: Optional[Union[dict, List[dict]]] = None,
    clear_cache=False,
    session_id=None,
) -> Dict[str, Any]:
    try:
        initialize_session_service()
        session_service = get_session_service()

        if clear_cache:
            session_service.clear_session(session_id)

        if session_id is None:
            session_id = session_service.generate_key(session_id=session_id, data_graph=data_graph)

        # Use async_to_sync to handle the asynchronous part of the session service
        session_data = async_to_sync(session_service.load_session, force_new_loop=True)(session_id, data_graph)
        logger.warning(f"session_data: {session_data}")
        graph, artifacts = session_data if session_data else (None, None)

        if not graph:
            raise ValueError("Graph not found in the session")

        # Use async_to_sync for the asynchronous build method
        built_object = async_to_sync(graph.build, force_new_loop=True)()

        logger.debug(f"Built object: {built_object}")

        processed_inputs = process_inputs(inputs, artifacts or {})
        result = async_to_sync(generate_result, force_new_loop=True)(built_object, processed_inputs)

        # Update the session with the new data
        session_service.update_session(session_id, (graph, artifacts))
        result_object = Result(result=result, session_id=session_id).model_dump()
        print(f"Result object: {result_object}")
        return result_object
    except Exception as e:
        logger.error(f"Error in process_graph_cached_task: {e}")
        # Handle the exception as needed, maybe re-raise or return an error message
        raise



from typing import TYPE_CHECKING

from langflow.services.credentials.service import CredentialService
from langflow.services.factory import ServiceFactory

if TYPE_CHECKING:
    from langflow.services.settings.service import SettingsService


class CredentialServiceFactory(ServiceFactory):
    def __init__(self):
        super().__init__(CredentialService)

    def create(self, settings_service: "SettingsService"):
        return CredentialService(settings_service)



from typing import Optional
from langflow import CustomComponent
from langchain.llms.base import BaseLanguageModel
from langchain_community.chat_models.azure_openai import AzureChatOpenAI


class AzureChatOpenAIComponent(CustomComponent):
    display_name: str = "AzureChatOpenAI"
    description: str = "LLM model from Azure OpenAI."
    documentation: str = "https://python.langchain.com/docs/integrations/llms/azure_openai"
    beta = False

    AZURE_OPENAI_MODELS = [
        "gpt-35-turbo",
        "gpt-35-turbo-16k",
        "gpt-35-turbo-instruct",
        "gpt-4",
        "gpt-4-32k",
        "gpt-4-vision",
    ]

    AZURE_OPENAI_API_VERSIONS = [
        "2023-03-15-preview",
        "2023-05-15",
        "2023-06-01-preview",
        "2023-07-01-preview",
        "2023-08-01-preview",
        "2023-09-01-preview",
        "2023-12-01-preview",
    ]

    def build_config(self):
        return {
            "model": {
                "display_name": "Model Name",
                "value": self.AZURE_OPENAI_MODELS[0],
                "options": self.AZURE_OPENAI_MODELS,
                "required": True,
            },
            "azure_endpoint": {
                "display_name": "Azure Endpoint",
                "required": True,
                "info": "Your Azure endpoint, including the resource.. Example: `https://example-resource.azure.openai.com/`",
            },
            "azure_deployment": {
                "display_name": "Deployment Name",
                "required": True,
            },
            "api_version": {
                "display_name": "API Version",
                "options": self.AZURE_OPENAI_API_VERSIONS,
                "value": self.AZURE_OPENAI_API_VERSIONS[-1],
                "required": True,
                "advanced": True,
            },
            "api_key": {"display_name": "API Key", "required": True, "password": True},
            "temperature": {
                "display_name": "Temperature",
                "value": 0.7,
                "field_type": "float",
                "required": False,
            },
            "max_tokens": {
                "display_name": "Max Tokens",
                "value": 1000,
                "required": False,
                "field_type": "int",
                "advanced": True,
                "info": "Maximum number of tokens to generate.",
            },
            "code": {"show": False},
        }

    def build(
        self,
        model: str,
        azure_endpoint: str,
        azure_deployment: str,
        api_key: str,
        api_version: str,
        temperature: float = 0.7,
        max_tokens: Optional[int] = 1000,
    ) -> BaseLanguageModel:
        try:
            llm = AzureChatOpenAI(
                model=model,
                azure_endpoint=azure_endpoint,
                azure_deployment=azure_deployment,
                api_version=api_version,
                api_key=api_key,
                temperature=temperature,
                max_tokens=max_tokens,
            )
        except Exception as e:
            raise ValueError("Could not connect to AzureOpenAI API.") from e
        return llm



import contextlib
import json
import os
from pathlib import Path
from typing import List, Optional

import yaml
from pydantic import model_validator, validator
from pydantic_settings import BaseSettings

from langflow.utils.logger import logger

BASE_COMPONENTS_PATH = str(Path(__file__).parent / "components")


class Settings(BaseSettings):
    CHAINS: dict = {}
    AGENTS: dict = {}
    PROMPTS: dict = {}
    LLMS: dict = {}
    TOOLS: dict = {}
    MEMORIES: dict = {}
    EMBEDDINGS: dict = {}
    VECTORSTORES: dict = {}
    DOCUMENTLOADERS: dict = {}
    WRAPPERS: dict = {}
    RETRIEVERS: dict = {}
    TOOLKITS: dict = {}
    TEXTSPLITTERS: dict = {}
    UTILITIES: dict = {}
    OUTPUT_PARSERS: dict = {}
    CUSTOM_COMPONENTS: dict = {}

    DEV: bool = False
    DATABASE_URL: Optional[str] = None
    CACHE: str = "InMemoryCache"
    REMOVE_API_KEYS: bool = False
    COMPONENTS_PATH: List[str] = []

    @validator("DATABASE_URL", pre=True)
    def set_database_url(cls, value):
        if not value:
            logger.debug("No database_url provided, trying LANGFLOW_DATABASE_URL env variable")
            if langflow_database_url := os.getenv("LANGFLOW_DATABASE_URL"):
                value = langflow_database_url
                logger.debug("Using LANGFLOW_DATABASE_URL env variable.")
            else:
                logger.debug("No DATABASE_URL env variable, using sqlite database")
                value = "sqlite:///./langflow.db"
        return value

    @validator("COMPONENTS_PATH", pre=True)
    def set_components_path(cls, value):
        if os.getenv("LANGFLOW_COMPONENTS_PATH"):
            logger.debug("Adding LANGFLOW_COMPONENTS_PATH to components_path")
            langflow_component_path = os.getenv("LANGFLOW_COMPONENTS_PATH")
            if Path(langflow_component_path).exists() and langflow_component_path not in value:
                if isinstance(langflow_component_path, list):
                    for path in langflow_component_path:
                        if path not in value:
                            value.append(path)
                    logger.debug(f"Extending {langflow_component_path} to components_path")
                elif langflow_component_path not in value:
                    value.append(langflow_component_path)
                    logger.debug(f"Appending {langflow_component_path} to components_path")

        if not value:
            value = [BASE_COMPONENTS_PATH]
            logger.debug("Setting default components path to components_path")
        elif BASE_COMPONENTS_PATH not in value:
            value.append(BASE_COMPONENTS_PATH)
            logger.debug("Adding default components path to components_path")

        logger.debug(f"Components path: {value}")
        return value

    class Config:
        validate_assignment = True
        extra = "ignore"
        env_prefix = "LANGFLOW_"

    @model_validator(mode="after")
    def validate_lists(cls, values):
        for key, value in values.items():
            if key != "dev" and not value:
                values[key] = []
        return values

    def update_from_yaml(self, file_path: str, dev: bool = False):
        new_settings = load_settings_from_yaml(file_path)
        self.CHAINS = new_settings.CHAINS or {}
        self.AGENTS = new_settings.AGENTS or {}
        self.PROMPTS = new_settings.PROMPTS or {}
        self.LLMS = new_settings.LLMS or {}
        self.TOOLS = new_settings.TOOLS or {}
        self.MEMORIES = new_settings.MEMORIES or {}
        self.WRAPPERS = new_settings.WRAPPERS or {}
        self.TOOLKITS = new_settings.TOOLKITS or {}
        self.TEXTSPLITTERS = new_settings.TEXTSPLITTERS or {}
        self.UTILITIES = new_settings.UTILITIES or {}
        self.EMBEDDINGS = new_settings.EMBEDDINGS or {}
        self.VECTORSTORES = new_settings.VECTORSTORES or {}
        self.DOCUMENTLOADERS = new_settings.DOCUMENTLOADERS or {}
        self.RETRIEVERS = new_settings.RETRIEVERS or {}
        self.OUTPUT_PARSERS = new_settings.OUTPUT_PARSERS or {}
        self.CUSTOM_COMPONENTS = new_settings.CUSTOM_COMPONENTS or {}
        self.COMPONENTS_PATH = new_settings.COMPONENTS_PATH or []
        self.DEV = dev

    def update_settings(self, **kwargs):
        logger.debug("Updating settings")
        for key, value in kwargs.items():
            # value may contain sensitive information, so we don't want to log it
            if not hasattr(self, key):
                logger.debug(f"Key {key} not found in settings")
                continue
            logger.debug(f"Updating {key}")
            if isinstance(getattr(self, key), list):
                # value might be a '[something]' string
                with contextlib.suppress(json.decoder.JSONDecodeError):
                    value = json.loads(str(value))
                if isinstance(value, list):
                    for item in value:
                        if isinstance(item, Path):
                            item = str(item)
                        if item not in getattr(self, key):
                            getattr(self, key).append(item)
                    logger.debug(f"Extended {key}")
                else:
                    if isinstance(value, Path):
                        value = str(value)
                    if value not in getattr(self, key):
                        getattr(self, key).append(value)
                        logger.debug(f"Appended {key}")

            else:
                setattr(self, key, value)
                logger.debug(f"Updated {key}")
            logger.debug(f"{key}: {getattr(self, key)}")


def save_settings_to_yaml(settings: Settings, file_path: str):
    with open(file_path, "w") as f:
        settings_dict = settings.model_dump()
        yaml.dump(settings_dict, f)


def load_settings_from_yaml(file_path: str) -> Settings:
    # Check if a string is a valid path or a file name
    if "/" not in file_path:
        # Get current path
        current_path = os.path.dirname(os.path.abspath(__file__))

        file_path = os.path.join(current_path, file_path)

    with open(file_path, "r") as f:
        settings_dict = yaml.safe_load(f)
        settings_dict = {k.upper(): v for k, v in settings_dict.items()}

        for key in settings_dict:
            if key not in Settings.model_fields.keys():
                raise KeyError(f"Key {key} not found in settings")
            logger.debug(f"Loading {len(settings_dict[key])} {key} from {file_path}")

    return Settings(**settings_dict)


settings = load_settings_from_yaml("config.yaml")



from langflow.services.factory import ServiceFactory
from langflow.services.auth.service import AuthService


class AuthServiceFactory(ServiceFactory):
    name = "auth_service"

    def __init__(self):
        super().__init__(AuthService)

    def create(self, settings_service):
        return AuthService(settings_service)



from typing import TYPE_CHECKING, Generator

from langflow.services import ServiceType, service_manager

if TYPE_CHECKING:
    from langflow.services.cache.service import BaseCacheService
    from langflow.services.chat.service import ChatService
    from langflow.services.credentials.service import CredentialService
    from langflow.services.database.service import DatabaseService
    from langflow.services.plugins.service import PluginService
    from langflow.services.session.service import SessionService
    from langflow.services.settings.service import SettingsService
    from langflow.services.store.service import StoreService
    from langflow.services.task.service import TaskService
    from sqlmodel import Session


def get_credential_service() -> "CredentialService":
    return service_manager.get(ServiceType.CREDENTIAL_SERVICE)  # type: ignore


def get_plugins_service() -> "PluginService":
    return service_manager.get(ServiceType.PLUGIN_SERVICE)  # type: ignore


def get_settings_service() -> "SettingsService":
    try:
        return service_manager.get(ServiceType.SETTINGS_SERVICE)  # type: ignore
    except ValueError:
        # initialize settings service
        from langflow.services.manager import initialize_settings_service

        initialize_settings_service()
        return service_manager.get(ServiceType.SETTINGS_SERVICE)  # type: ignore


def get_db_service() -> "DatabaseService":
    return service_manager.get(ServiceType.DATABASE_SERVICE)  # type: ignore


def get_session() -> Generator["Session", None, None]:
    db_service = get_db_service()
    yield from db_service.get_session()


def get_cache_service() -> "BaseCacheService":
    return service_manager.get(ServiceType.CACHE_SERVICE)  # type: ignore


def get_session_service() -> "SessionService":
    return service_manager.get(ServiceType.SESSION_SERVICE)  # type: ignore


def get_task_service() -> "TaskService":
    return service_manager.get(ServiceType.TASK_SERVICE)  # type: ignore


def get_chat_service() -> "ChatService":
    return service_manager.get(ServiceType.CHAT_SERVICE)  # type: ignore


def get_store_service() -> "StoreService":
    return service_manager.get(ServiceType.STORE_SERVICE)  # type: ignore



from typing import Any

from .constants import (
    AgentExecutor,
    BaseChatMemory,
    BaseLanguageModel,
    BaseLLM,
    BaseLoader,
    BaseMemory,
    BaseOutputParser,
    BasePromptTemplate,
    BaseRetriever,
    Callable,
    Chain,
    ChatPromptTemplate,
    Data,
    Document,
    Embeddings,
    NestedDict,
    Object,
    Prompt,
    PromptTemplate,
    TextSplitter,
    Tool,
    VectorStore,
)
from .range_spec import RangeSpec


def _import_template_field():
    from langflow.template.field.base import TemplateField

    return TemplateField


def __getattr__(name: str) -> Any:
    # This is to avoid circular imports
    if name == "TemplateField":
        return _import_template_field()
    elif name == "RangeSpec":
        return RangeSpec
    # The other names should work as if they were imported from constants
    # Import the constants module langflow.field_typing.constants
    from . import constants

    return getattr(constants, name)


__all__ = [
    "NestedDict",
    "Data",
    "Tool",
    "PromptTemplate",
    "Chain",
    "BaseChatMemory",
    "BaseLLM",
    "BaseLanguageModel",
    "BaseLoader",
    "BaseMemory",
    "BaseOutputParser",
    "BaseRetriever",
    "VectorStore",
    "Embeddings",
    "TextSplitter",
    "Document",
    "AgentExecutor",
    "Object",
    "Callable",
    "BasePromptTemplate",
    "ChatPromptTemplate",
    "Prompt",
    "RangeSpec",
    "TemplateField",
]



from contextlib import contextmanager
from dataclasses import dataclass
from typing import TYPE_CHECKING

from alembic.util.exc import CommandError
from loguru import logger
from sqlmodel import Session, text

if TYPE_CHECKING:
    from langflow.services.database.service import DatabaseService


def initialize_database(fix_migration: bool = False):
    logger.debug("Initializing database")
    from langflow.services.deps import get_db_service

    database_service: "DatabaseService" = get_db_service()
    try:
        database_service.create_db_and_tables()
    except Exception as exc:
        # if the exception involves tables already existing
        # we can ignore it
        if "already exists" not in str(exc):
            logger.error(f"Error creating DB and tables: {exc}")
            raise RuntimeError("Error creating DB and tables") from exc
    try:
        database_service.check_schema_health()
    except Exception as exc:
        logger.error(f"Error checking schema health: {exc}")
        raise RuntimeError("Error checking schema health") from exc
    try:
        database_service.run_migrations(fix=fix_migration)
    except CommandError as exc:
        # if "overlaps with other requested revisions" or "Can't locate revision identified by"
        # are not in the exception, we can't handle it
        if "overlaps with other requested revisions" not in str(
            exc
        ) and "Can't locate revision identified by" not in str(exc):
            raise exc
        # This means there's wrong revision in the DB
        # We need to delete the alembic_version table
        # and run the migrations again
        logger.warning("Wrong revision in DB, deleting alembic_version table and running migrations again")
        with session_getter(database_service) as session:
            session.exec(text("DROP TABLE alembic_version"))
        database_service.run_migrations(fix=fix_migration)
    except Exception as exc:
        # if the exception involves tables already existing
        # we can ignore it
        if "already exists" not in str(exc):
            logger.error(exc)
        raise exc
    logger.debug("Database initialized")


@contextmanager
def session_getter(db_service: "DatabaseService"):
    try:
        session = Session(db_service.engine)
        yield session
    except Exception as e:
        print("Session rollback because of exception:", e)
        session.rollback()
        raise
    finally:
        session.close()


@dataclass
class Result:
    name: str
    type: str
    success: bool


@dataclass
class TableResults:
    table_name: str
    results: list[Result]



from typing import Optional, List, Dict, Any
from langflow import CustomComponent
from langchain_community.llms.llamacpp import LlamaCpp


class LlamaCppComponent(CustomComponent):
    display_name = "LlamaCpp"
    description = "llama.cpp model."
    documentation = "https://python.langchain.com/docs/modules/model_io/models/llms/integrations/llamacpp"

    def build_config(self):
        return {
            "grammar": {"display_name": "Grammar", "advanced": True},
            "cache": {"display_name": "Cache", "advanced": True},
            "client": {"display_name": "Client", "advanced": True},
            "echo": {"display_name": "Echo", "advanced": True},
            "f16_kv": {"display_name": "F16 KV", "advanced": True},
            "grammar_path": {"display_name": "Grammar Path", "advanced": True},
            "last_n_tokens_size": {"display_name": "Last N Tokens Size", "advanced": True},
            "logits_all": {"display_name": "Logits All", "advanced": True},
            "logprobs": {"display_name": "Logprobs", "advanced": True},
            "lora_base": {"display_name": "Lora Base", "advanced": True},
            "lora_path": {"display_name": "Lora Path", "advanced": True},
            "max_tokens": {"display_name": "Max Tokens", "advanced": True},
            "metadata": {"display_name": "Metadata", "advanced": True},
            "model_kwargs": {"display_name": "Model Kwargs", "advanced": True},
            "model_path": {
                "display_name": "Model Path",
                "field_type": "file",
                "file_types": [".bin"],
                "required": True,
            },
            "n_batch": {"display_name": "N Batch", "advanced": True},
            "n_ctx": {"display_name": "N Ctx", "advanced": True},
            "n_gpu_layers": {"display_name": "N GPU Layers", "advanced": True},
            "n_parts": {"display_name": "N Parts", "advanced": True},
            "n_threads": {"display_name": "N Threads", "advanced": True},
            "repeat_penalty": {"display_name": "Repeat Penalty", "advanced": True},
            "rope_freq_base": {"display_name": "Rope Freq Base", "advanced": True},
            "rope_freq_scale": {"display_name": "Rope Freq Scale", "advanced": True},
            "seed": {"display_name": "Seed", "advanced": True},
            "stop": {"display_name": "Stop", "advanced": True},
            "streaming": {"display_name": "Streaming", "advanced": True},
            "suffix": {"display_name": "Suffix", "advanced": True},
            "tags": {"display_name": "Tags", "advanced": True},
            "temperature": {"display_name": "Temperature"},
            "top_k": {"display_name": "Top K", "advanced": True},
            "top_p": {"display_name": "Top P", "advanced": True},
            "use_mlock": {"display_name": "Use Mlock", "advanced": True},
            "use_mmap": {"display_name": "Use Mmap", "advanced": True},
            "verbose": {"display_name": "Verbose", "advanced": True},
            "vocab_only": {"display_name": "Vocab Only", "advanced": True},
        }

    def build(
        self,
        model_path: str,
        grammar: Optional[str] = None,
        cache: Optional[bool] = None,
        client: Optional[Any] = None,
        echo: Optional[bool] = False,
        f16_kv: bool = True,
        grammar_path: Optional[str] = None,
        last_n_tokens_size: Optional[int] = 64,
        logits_all: bool = False,
        logprobs: Optional[int] = None,
        lora_base: Optional[str] = None,
        lora_path: Optional[str] = None,
        max_tokens: Optional[int] = 256,
        metadata: Optional[Dict] = None,
        model_kwargs: Dict = {},
        n_batch: Optional[int] = 8,
        n_ctx: int = 512,
        n_gpu_layers: Optional[int] = 1,
        n_parts: int = -1,
        n_threads: Optional[int] = 1,
        repeat_penalty: Optional[float] = 1.1,
        rope_freq_base: float = 10000.0,
        rope_freq_scale: float = 1.0,
        seed: int = -1,
        stop: Optional[List[str]] = [],
        streaming: bool = True,
        suffix: Optional[str] = "",
        tags: Optional[List[str]] = [],
        temperature: Optional[float] = 0.8,
        top_k: Optional[int] = 40,
        top_p: Optional[float] = 0.95,
        use_mlock: bool = False,
        use_mmap: Optional[bool] = True,
        verbose: bool = True,
        vocab_only: bool = False,
    ) -> LlamaCpp:
        return LlamaCpp(
            model_path=model_path,
            grammar=grammar,
            cache=cache,
            client=client,
            echo=echo,
            f16_kv=f16_kv,
            grammar_path=grammar_path,
            last_n_tokens_size=last_n_tokens_size,
            logits_all=logits_all,
            logprobs=logprobs,
            lora_base=lora_base,
            lora_path=lora_path,
            max_tokens=max_tokens,
            metadata=metadata,
            model_kwargs=model_kwargs,
            n_batch=n_batch,
            n_ctx=n_ctx,
            n_gpu_layers=n_gpu_layers,
            n_parts=n_parts,
            n_threads=n_threads,
            repeat_penalty=repeat_penalty,
            rope_freq_base=rope_freq_base,
            rope_freq_scale=rope_freq_scale,
            seed=seed,
            stop=stop,
            streaming=streaming,
            suffix=suffix,
            tags=tags,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            use_mlock=use_mlock,
            use_mmap=use_mmap,
            verbose=verbose,
            vocab_only=vocab_only,
        )



from typing import ClassVar, Dict, Optional

from langflow.template.field.base import TemplateField
from langflow.template.frontend_node.base import FrontendNode


def build_file_field(fileTypes: list, name: str = "file_path") -> TemplateField:
    """Build a template field for a document loader."""
    return TemplateField(
        field_type="file",
        required=True,
        show=True,
        name=name,
        value="",
        file_types=fileTypes,
    )


class DocumentLoaderFrontNode(FrontendNode):
    def add_extra_base_classes(self) -> None:
        self.base_classes = ["Document"]
        self.output_types = ["Document"]

    file_path_templates: ClassVar[Dict] = {
        "AirbyteJSONLoader": build_file_field(
            fileTypes=[".json"],
        ),
        "CoNLLULoader": build_file_field(
            fileTypes=[".csv"],
        ),
        "CSVLoader": build_file_field(
            fileTypes=[".csv"],
        ),
        "UnstructuredEmailLoader": build_file_field(
            fileTypes=[".eml"],
        ),
        "EverNoteLoader": build_file_field(
            fileTypes=[".xml"],
        ),
        "FacebookChatLoader": build_file_field(
            fileTypes=[".json"],
        ),
        "BSHTMLLoader": build_file_field(
            fileTypes=[".html"],
        ),
        "UnstructuredHTMLLoader": build_file_field(fileTypes=[".html"]),
        "UnstructuredImageLoader": build_file_field(
            fileTypes=[".jpg", ".jpeg", ".png", ".gif", ".bmp"],
        ),
        "UnstructuredMarkdownLoader": build_file_field(
            fileTypes=[".md"],
        ),
        "PyPDFLoader": build_file_field(
            fileTypes=[".pdf"],
        ),
        "UnstructuredPowerPointLoader": build_file_field(
            fileTypes=[".pptx", ".ppt"],
        ),
        "SRTLoader": build_file_field(
            fileTypes=[".srt"],
        ),
        "TelegramChatLoader": build_file_field(
            fileTypes=[".json"],
        ),
        "TextLoader": build_file_field(
            fileTypes=[".txt"],
        ),
        "UnstructuredWordDocumentLoader": build_file_field(
            fileTypes=[".docx", ".doc"],
        ),
    }

    def add_extra_fields(self) -> None:
        name = None
        display_name = "Web Page"
        if self.template.type_name in {"GitLoader"}:
            # Add fields repo_path, clone_url, branch and file_filter
            self.template.add_field(
                TemplateField(
                    field_type="str",
                    required=True,
                    show=True,
                    name="repo_path",
                    value="",
                    display_name="Path to repository",
                    advanced=False,
                )
            )
            self.template.add_field(
                TemplateField(
                    field_type="str",
                    required=False,
                    show=True,
                    name="clone_url",
                    value="",
                    display_name="Clone URL",
                    advanced=False,
                )
            )
            self.template.add_field(
                TemplateField(
                    field_type="str",
                    required=True,
                    show=True,
                    name="branch",
                    value="",
                    display_name="Branch",
                    advanced=False,
                )
            )
            self.template.add_field(
                TemplateField(
                    field_type="str",
                    required=False,
                    show=True,
                    name="file_filter",
                    value="",
                    display_name="File extensions (comma-separated)",
                    advanced=False,
                )
            )
        elif self.template.type_name in {"SlackDirectoryLoader"}:
            self.template.add_field(
                TemplateField(
                    field_type="file",
                    required=True,
                    show=True,
                    name="zip_path",
                    value="",
                    display_name="Path to zip file",
                    file_types=[".zip"],
                )
            )
            self.template.add_field(
                TemplateField(
                    field_type="str",
                    required=False,
                    show=True,
                    name="workspace_url",
                    value="",
                    display_name="Workspace URL",
                    advanced=False,
                )
            )
        elif self.template.type_name in self.file_path_templates:
            self.template.add_field(self.file_path_templates[self.template.type_name])
        elif self.template.type_name in {
            "WebBaseLoader",
            "AZLyricsLoader",
            "CollegeConfidentialLoader",
            "HNLoader",
            "IFixitLoader",
            "IMSDbLoader",
            "GutenbergLoader",
        }:
            name = "web_path"
        elif self.template.type_name in {"GutenbergLoader"}:
            name = "file_path"
        elif self.template.type_name in {"GitbookLoader"}:
            name = "web_page"
        elif self.template.type_name in {
            "DirectoryLoader",
            "ReadTheDocsLoader",
            "NotionDirectoryLoader",
            "PyPDFDirectoryLoader",
        }:
            name = "path"
            display_name = "Local directory"
        if name:
            if self.template.type_name in {"DirectoryLoader"}:
                for field in build_directory_loader_fields():
                    self.template.add_field(field)
            else:
                self.template.add_field(
                    TemplateField(
                        field_type="str",
                        required=True,
                        show=True,
                        name=name,
                        value="",
                        display_name=display_name,
                    )
                )
            # add a metadata field of type dict
        self.template.add_field(
            TemplateField(
                field_type="dict",
                required=False,
                show=True,
                name="metadata",
                value={},
                display_name="Metadata",
                multiline=False,
            )
        )

    @staticmethod
    def format_field(field: TemplateField, name: Optional[str] = None) -> None:
        FrontendNode.format_field(field, name)
        if field.name == "metadata":
            field.show = True
            field.advanced = False
        field.show = True


def build_directory_loader_fields():
    # if loader_kwargs is None:
    #         loader_kwargs = {}
    # self.path = path
    # self.glob = glob
    # self.load_hidden = load_hidden
    # self.loader_cls = loader_cls
    # self.loader_kwargs = loader_kwargs
    # self.silent_errors = silent_errors
    # self.recursive = recursive
    # self.show_progress = show_progress
    # self.use_multithreading = use_multithreading
    # self.max_concurrency = max_concurrency
    # Based on the above fields, we can build the following fields:
    # path, glob, load_hidden, silent_errors, recursive, show_progress, use_multithreading, max_concurrency
    # path
    path = TemplateField(
        field_type="str",
        required=True,
        show=True,
        name="path",
        value="",
        display_name="Local directory",
        advanced=False,
    )
    # glob
    glob = TemplateField(
        field_type="str",
        required=True,
        show=True,
        name="glob",
        value="**/*.txt",
        display_name="glob",
        advanced=False,
    )
    # load_hidden
    load_hidden = TemplateField(
        field_type="bool",
        required=False,
        show=True,
        name="load_hidden",
        value="False",
        display_name="Load hidden files",
        advanced=True,
    )
    # silent_errors
    silent_errors = TemplateField(
        field_type="bool",
        required=False,
        show=True,
        name="silent_errors",
        value="False",
        display_name="Silent errors",
        advanced=True,
    )
    # recursive
    recursive = TemplateField(
        field_type="bool",
        required=False,
        show=True,
        name="recursive",
        value="True",
        display_name="Recursive",
        advanced=True,
    )

    # use_multithreading
    use_multithreading = TemplateField(
        field_type="bool",
        required=False,
        show=True,
        name="use_multithreading",
        value="True",
        display_name="Use multithreading",
        advanced=True,
    )
    # max_concurrency
    max_concurrency = TemplateField(
        field_type="int",
        required=False,
        show=True,
        name="max_concurrency",
        value=10,
        display_name="Max concurrency",
        advanced=True,
    )

    return (
        path,
        glob,
        load_hidden,
        silent_errors,
        recursive,
        use_multithreading,
        max_concurrency,
    )



from langflow.services.chat.service import ChatService
from langflow.services.factory import ServiceFactory


class ChatServiceFactory(ServiceFactory):
    def __init__(self):
        super().__init__(ChatService)

    def create(self):
        # Here you would have logic to create and configure a ChatService
        return ChatService()



from typing import TYPE_CHECKING
from langflow.services.store.service import StoreService
from langflow.services.factory import ServiceFactory

if TYPE_CHECKING:
    from langflow.services.settings.service import SettingsService


class StoreServiceFactory(ServiceFactory):
    def __init__(self):
        super().__init__(StoreService)

    def create(self, settings_service: "SettingsService"):
        return StoreService(settings_service)



from langflow import CustomComponent
from typing import Optional, Dict
from langchain_community.utilities.searx_search import SearxSearchWrapper


class SearxSearchWrapperComponent(CustomComponent):
    display_name = "SearxSearchWrapper"
    description = "Wrapper for Searx API."

    def build_config(self):
        return {
            "headers": {
                "field_type": "dict",
                "display_name": "Headers",
                "multiline": True,
                "value": '{"Authorization": "Bearer <token>"}',
            },
            "k": {"display_name": "k", "advanced": True, "field_type": "int", "value": 10},
            "searx_host": {
                "display_name": "Searx Host",
                "field_type": "str",
                "value": "https://searx.example.com",
                "advanced": True,
            },
        }

    def build(
        self,
        k: int = 10,
        headers: Optional[Dict[str, str]] = None,
        searx_host: str = "https://searx.example.com",
    ) -> SearxSearchWrapper:
        return SearxSearchWrapper(headers=headers, k=k, searx_host=searx_host)



import warnings
from pathlib import Path
from typing import TYPE_CHECKING, List

from fastapi import HTTPException
from platformdirs import user_cache_dir

from langflow.services.store.schema import StoreComponentCreate
from langflow.services.store.utils import get_lf_version_from_pypi

if TYPE_CHECKING:
    from langflow.services.database.models.flow.model import Flow


API_WORDS = ["api", "key", "token"]


def has_api_terms(word: str):
    return "api" in word and ("key" in word or ("token" in word and "tokens" not in word))


def remove_api_keys(flow: dict):
    """Remove api keys from flow data."""
    if flow.get("data") and flow["data"].get("nodes"):
        for node in flow["data"]["nodes"]:
            node_data = node.get("data").get("node")
            template = node_data.get("template")
            for value in template.values():
                if isinstance(value, dict) and has_api_terms(value["name"]) and value.get("password"):
                    value["value"] = None

    return flow


def build_input_keys_response(langchain_object, artifacts):
    """Build the input keys response."""

    input_keys_response = {
        "input_keys": {key: "" for key in langchain_object.input_keys},
        "memory_keys": [],
        "handle_keys": artifacts.get("handle_keys", []),
    }

    # Set the input keys values from artifacts
    for key, value in artifacts.items():
        if key in input_keys_response["input_keys"]:
            input_keys_response["input_keys"][key] = value
    # If the object has memory, that memory will have a memory_variables attribute
    # memory variables should be removed from the input keys
    if hasattr(langchain_object, "memory") and hasattr(langchain_object.memory, "memory_variables"):
        # Remove memory variables from input keys
        input_keys_response["input_keys"] = {
            key: value
            for key, value in input_keys_response["input_keys"].items()
            if key not in langchain_object.memory.memory_variables
        }
        # Add memory variables to memory_keys
        input_keys_response["memory_keys"] = langchain_object.memory.memory_variables

    if hasattr(langchain_object, "prompt") and hasattr(langchain_object.prompt, "template"):
        input_keys_response["template"] = langchain_object.prompt.template

    return input_keys_response


def update_frontend_node_with_template_values(frontend_node, raw_frontend_node):
    """
    Updates the given frontend node with values from the raw template data.

    :param frontend_node: A dict representing a built frontend node.
    :param raw_template_data: A dict representing raw template data.
    :return: Updated frontend node.
    """
    if not is_valid_data(frontend_node, raw_frontend_node):
        return frontend_node

    # Check if the display_name is different than "CustomComponent"
    # if so, update the display_name in the frontend_node
    if raw_frontend_node["display_name"] != "CustomComponent":
        frontend_node["display_name"] = raw_frontend_node["display_name"]

    update_template_values(frontend_node["template"], raw_frontend_node["template"])

    return frontend_node


def raw_frontend_data_is_valid(raw_frontend_data):
    """Check if the raw frontend data is valid for processing."""
    return "template" in raw_frontend_data and "display_name" in raw_frontend_data


def is_valid_data(frontend_node, raw_frontend_data):
    """Check if the data is valid for processing."""

    return frontend_node and "template" in frontend_node and raw_frontend_data_is_valid(raw_frontend_data)


def update_template_values(frontend_template, raw_template):
    """Updates the frontend template with values from the raw template."""
    for key, value_dict in raw_template.items():
        if key == "code" or not isinstance(value_dict, dict):
            continue

        update_template_field(frontend_template, key, value_dict)


def update_template_field(frontend_template, key, value_dict):
    """Updates a specific field in the frontend template."""
    template_field = frontend_template.get(key)
    if not template_field or template_field.get("type") != value_dict.get("type"):
        return

    if "value" in value_dict and value_dict["value"]:
        template_field["value"] = value_dict["value"]

    if "file_path" in value_dict and value_dict["file_path"]:
        file_path_value = get_file_path_value(value_dict["file_path"])
        if not file_path_value:
            # If the file does not exist, remove the value from the template_field["value"]
            template_field["value"] = ""
        template_field["file_path"] = file_path_value


def get_file_path_value(file_path):
    """Get the file path value if the file exists, else return empty string."""
    try:
        path = Path(file_path)
    except TypeError:
        return ""

    # Check for safety
    # If the path is not in the cache dir, return empty string
    # This is to prevent access to files outside the cache dir
    # If the path is not a file, return empty string
    if not path.exists() or not str(path).startswith(user_cache_dir("langflow", "langflow")):
        return ""
    return file_path


def validate_is_component(flows: List["Flow"]):
    for flow in flows:
        if not flow.data or flow.is_component is not None:
            continue

        is_component = get_is_component_from_data(flow.data)
        if is_component is not None:
            flow.is_component = is_component
        else:
            flow.is_component = len(flow.data.get("nodes", [])) == 1
    return flows


def get_is_component_from_data(data: dict):
    """Returns True if the data is a component."""
    return data.get("is_component")


async def check_langflow_version(component: StoreComponentCreate):
    from langflow import __version__ as current_version

    if not component.last_tested_version:
        component.last_tested_version = current_version

    langflow_version = get_lf_version_from_pypi()
    if langflow_version is None:
        raise HTTPException(status_code=500, detail="Unable to verify the latest version of Langflow")
    elif langflow_version != component.last_tested_version:
        warnings.warn(
            f"Your version of Langflow ({component.last_tested_version}) is outdated. "
            f"Please update to the latest version ({langflow_version}) and try again."
        )


def format_elapsed_time(elapsed_time) -> str:
    # Format elapsed time to human readable format coming from
    # perf_counter()
    # If the elapsed time is less than 1 second, return ms
    # If the elapsed time is less than 1 minute, return seconds rounded to 2 decimals
    time_str = ""
    if elapsed_time < 1:
        elapsed_time = int(round(elapsed_time * 1000))
        time_str = f"{elapsed_time} ms"
    elif elapsed_time < 60:
        elapsed_time = round(elapsed_time, 2)
        time_str = f"{elapsed_time} seconds"
    else:
        elapsed_time = round(elapsed_time / 60, 2)
        time_str = f"{elapsed_time} minutes"
    return time_str



from langflow import CustomComponent
from langchain.agents.agent_toolkits.vectorstore.toolkit import VectorStoreToolkit
from langchain.agents.agent_toolkits.vectorstore.toolkit import VectorStoreInfo
from langflow.field_typing import (
    BaseLanguageModel,
)
from langflow.field_typing import (
    Tool,
)
from typing import Union


class VectorStoreToolkitComponent(CustomComponent):
    display_name = "VectorStoreToolkit"
    description = "Toolkit for interacting with a Vector Store."

    def build_config(self):
        return {
            "vectorstore_info": {"display_name": "Vector Store Info"},
            "llm": {"display_name": "LLM"},
        }

    def build(
        self,
        vectorstore_info: VectorStoreInfo,
        llm: BaseLanguageModel,
    ) -> Union[Tool, VectorStoreToolkit]:
        return VectorStoreToolkit(vectorstore_info=vectorstore_info, llm=llm)



import importlib
import inspect
import os
from typing import TYPE_CHECKING, Union

from langflow.services.base import Service
from langflow.services.plugins.base import BasePlugin, CallbackPlugin
from loguru import logger

if TYPE_CHECKING:
    from langflow.services.settings.service import SettingsService


class PluginService(Service):
    name = "plugin_service"

    def __init__(self, settings_service: "SettingsService"):
        self.plugins: dict[str, BasePlugin] = {}
        # plugin_dir = settings_service.settings.PLUGIN_DIR
        self.plugin_dir = os.path.dirname(__file__)
        self.plugins_base_module = "langflow.services.plugins"
        self.load_plugins()

    def load_plugins(self):
        base_files = ["base.py", "service.py", "factory.py", "__init__.py"]
        for module in os.listdir(self.plugin_dir):
            if module.endswith(".py") and module not in base_files:
                plugin_name = module[:-3]
                module_path = f"{self.plugins_base_module}.{plugin_name}"
                try:
                    mod = importlib.import_module(module_path)
                    for attr_name in dir(mod):
                        attr = getattr(mod, attr_name)
                        if (
                            inspect.isclass(attr)
                            and issubclass(attr, BasePlugin)
                            and attr not in [CallbackPlugin, BasePlugin]
                        ):
                            self.register_plugin(plugin_name, attr())
                except Exception as exc:
                    logger.error(f"Error loading plugin {plugin_name}: {exc}")

    def register_plugin(self, plugin_name, plugin_instance):
        self.plugins[plugin_name] = plugin_instance
        plugin_instance.initialize()

    def get_plugin(self, plugin_name) -> Union[BasePlugin, None]:
        return self.plugins.get(plugin_name)

    def get(self, plugin_name):
        if plugin := self.get_plugin(plugin_name):
            return plugin.get()
        return None

    def teardown(self):
        for plugin in self.plugins.values():
            plugin.teardown()

    def get_callbacks(self, _id=None):
        callbacks = []
        for plugin in self.plugins.values():
            if isinstance(plugin, CallbackPlugin):
                callback = plugin.get_callback(_id=_id)
                if callback:
                    callbacks.append(callback)
        return callbacks



from langflow import CustomComponent
from langchain.agents import AgentExecutor, create_json_agent
from langflow.field_typing import (
    BaseLanguageModel,
)
from langchain_community.agent_toolkits.json.toolkit import JsonToolkit


class JsonAgentComponent(CustomComponent):
    display_name = "JsonAgent"
    description = "Construct a json agent from an LLM and tools."

    def build_config(self):
        return {
            "llm": {"display_name": "LLM"},
            "toolkit": {"display_name": "Toolkit"},
        }

    def build(
        self,
        llm: BaseLanguageModel,
        toolkit: JsonToolkit,
    ) -> AgentExecutor:
        return create_json_agent(llm=llm, toolkit=toolkit)



from langflow import CustomComponent
from langchain.chains import ConversationChain
from typing import Optional, Union, Callable
from langflow.field_typing import BaseLanguageModel, BaseMemory, Chain


class ConversationChainComponent(CustomComponent):
    display_name = "ConversationChain"
    description = "Chain to have a conversation and load context from memory."

    def build_config(self):
        return {
            "prompt": {"display_name": "Prompt"},
            "llm": {"display_name": "LLM"},
            "memory": {
                "display_name": "Memory",
                "info": "Memory to load context from. If none is provided, a ConversationBufferMemory will be used.",
            },
            "code": {"show": False},
        }

    def build(
        self,
        llm: BaseLanguageModel,
        memory: Optional[BaseMemory] = None,
    ) -> Union[Chain, Callable]:
        if memory is None:
            return ConversationChain(llm=llm)
        return ConversationChain(llm=llm, memory=memory)



from typing import Dict

# Assuming the existence of GoogleSerperAPIWrapper class in the serper module
# If this class does not exist, you would need to create it or import the appropriate class from another module
from langchain_community.utilities.google_serper import GoogleSerperAPIWrapper
from langflow import CustomComponent


class GoogleSerperAPIWrapperComponent(CustomComponent):
    display_name = "GoogleSerperAPIWrapper"
    description = "Wrapper around the Serper.dev Google Search API."

    def build_config(self) -> Dict[str, Dict]:
        return {
            "result_key_for_type": {
                "display_name": "Result Key for Type",
                "show": True,
                "multiline": False,
                "password": False,
                "name": "result_key_for_type",
                "advanced": False,
                "dynamic": False,
                "info": "",
                "field_type": "dict",
                "list": False,
                "value": {"news": "news", "places": "places", "images": "images", "search": "organic"},
            },
            "serper_api_key": {
                "display_name": "Serper API Key",
                "show": True,
                "multiline": False,
                "password": True,
                "name": "serper_api_key",
                "advanced": False,
                "dynamic": False,
                "info": "",
                "type": "str",
                "list": False,
            },
        }

    def build(
        self,
        serper_api_key: str,
    ) -> GoogleSerperAPIWrapper:
        return GoogleSerperAPIWrapper(serper_api_key=serper_api_key)



from langflow.services.base import Service
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from langflow.services.settings.service import SettingsService


class AuthService(Service):
    name = "auth_service"

    def __init__(self, settings_service: "SettingsService"):
        self.settings_service = settings_service



### JSON Document Builder

# Build a Document containing a JSON object using a key and another Document page content.

# **Params**

# - **Key:** The key to use for the JSON object.
# - **Document:** The Document page to use for the JSON object.

# **Output**

# - **Document:** The Document containing the JSON object.

from langchain_core.documents import Document
from langflow import CustomComponent
from langflow.services.database.models.base import orjson_dumps


class JSONDocumentBuilder(CustomComponent):
    display_name: str = "JSON Document Builder"
    description: str = "Build a Document containing a JSON object using a key and another Document page content."
    output_types: list[str] = ["Document"]
    beta = True
    documentation: str = "https://docs.langflow.org/components/utilities#json-document-builder"

    field_config = {
        "key": {"display_name": "Key"},
        "document": {"display_name": "Document"},
    }

    def build(
        self,
        key: str,
        document: Document,
    ) -> Document:
        documents = None
        if isinstance(document, list):
            documents = [
                Document(page_content=orjson_dumps({key: doc.page_content}, indent_2=False)) for doc in document
            ]
        elif isinstance(document, Document):
            documents = Document(page_content=orjson_dumps({key: document.page_content}, indent_2=False))
        else:
            raise TypeError(f"Expected Document or list of Documents, got {type(document)}")
        self.repr_value = documents
        return documents



from langflow import CustomComponent
from langflow.field_typing import Data


class Component(CustomComponent):
    documentation: str = "http://docs.langflow.org/components/custom"

    def build_config(self):
        return {"param": {"display_name": "Parameter"}}

    def build(self, param: Data) -> Data:
        return param



from pydantic import BaseModel, field_validator


class RangeSpec(BaseModel):
    min: float = -1.0
    max: float = 1.0
    step: float = 0.1

    @field_validator("max")
    @classmethod
    def max_must_be_greater_than_min(cls, v, values, **kwargs):
        if "min" in values.data and v <= values.data["min"]:
            raise ValueError("max must be greater than min")
        return v

    @field_validator("step")
    @classmethod
    def step_must_be_positive(cls, v):
        if v <= 0:
            raise ValueError("step must be positive")
        return v



import ast
import inspect
import types
from typing import TYPE_CHECKING, Any, Coroutine, Dict, List, Optional

from langflow.graph.utils import UnbuiltObject
from langflow.interface.initialize import loading
from langflow.interface.listing import lazy_load_dict
from langflow.utils.constants import DIRECT_TYPES
from langflow.utils.util import sync_to_async
from loguru import logger

if TYPE_CHECKING:
    from langflow.graph.edge.base import Edge
    from langflow.graph.graph.base import Graph


class Vertex:
    def __init__(
        self,
        data: Dict,
        graph: "Graph",
        base_type: Optional[str] = None,
        is_task: bool = False,
        params: Optional[Dict] = None,
    ) -> None:
        self.graph = graph
        self.id: str = data["id"]
        self._data = data
        self.base_type: Optional[str] = base_type
        self._parse_data()
        self._built_object = UnbuiltObject()
        self._built = False
        self.artifacts: Dict[str, Any] = {}
        self.task_id: Optional[str] = None
        self.is_task = is_task
        self.params = params or {}
        self.parent_node_id: Optional[str] = self._data.get("parent_node_id")
        self.parent_is_top_level = False

    @property
    def edges(self) -> List["Edge"]:
        return self.graph.get_vertex_edges(self.id)

    def __getstate__(self):
        return {
            "_data": self._data,
            "params": {},
            "base_type": self.base_type,
            "is_task": self.is_task,
            "id": self.id,
            "_built_object": UnbuiltObject(),
            "_built": False,
            "parent_node_id": self.parent_node_id,
            "parent_is_top_level": self.parent_is_top_level,
        }

    def __setstate__(self, state):
        self._data = state["_data"]
        self.params = state["params"]
        self.base_type = state["base_type"]
        self.is_task = state["is_task"]
        self.id = state["id"]
        self._parse_data()
        if "_built_object" in state:
            self._built_object = state["_built_object"]
            self._built = state["_built"]
        else:
            self._built_object = UnbuiltObject()
            self._built = False
        self.artifacts: Dict[str, Any] = {}
        self.task_id: Optional[str] = None
        self.parent_node_id = state["parent_node_id"]
        self.parent_is_top_level = state["parent_is_top_level"]

    def set_top_level(self, top_level_vertices: List[str]) -> None:
        self.parent_is_top_level = self.parent_node_id in top_level_vertices

    def _parse_data(self) -> None:
        self.data = self._data["data"]
        self.output = self.data["node"]["base_classes"]
        template_dicts = {key: value for key, value in self.data["node"]["template"].items() if isinstance(value, dict)}

        self.required_inputs = [
            template_dicts[key]["type"] for key, value in template_dicts.items() if value["required"]
        ]
        self.optional_inputs = [
            template_dicts[key]["type"] for key, value in template_dicts.items() if not value["required"]
        ]
        # Add the template_dicts[key]["input_types"] to the optional_inputs
        self.optional_inputs.extend(
            [input_type for value in template_dicts.values() for input_type in value.get("input_types", [])]
        )

        template_dict = self.data["node"]["template"]
        self.vertex_type = (
            self.data["type"]
            if "Tool" not in self.output or template_dict["_type"].islower()
            else template_dict["_type"]
        )

        if self.base_type is None:
            for base_type, value in lazy_load_dict.ALL_TYPES_DICT.items():
                if self.vertex_type in value:
                    self.base_type = base_type
                    break

    def get_task(self):
        # using the task_id, get the task from celery
        # and return it
        from celery.result import AsyncResult  # type: ignore

        return AsyncResult(self.task_id)

    def _build_params(self):
        # sourcery skip: merge-list-append, remove-redundant-if
        # Some params are required, some are optional
        # but most importantly, some params are python base classes
        # like str and others are LangChain objects like LLMChain, BasePromptTemplate
        # so we need to be able to distinguish between the two

        # The dicts with "type" == "str" are the ones that are python base classes
        # and most likely have a "value" key

        # So for each key besides "_type" in the template dict, we have a dict
        # with a "type" key. If the type is not "str", then we need to get the
        # edge that connects to that node and get the Node with the required data
        # and use that as the value for the param
        # If the type is "str", then we need to get the value of the "value" key
        # and use that as the value for the param

        if self.graph is None:
            raise ValueError("Graph not found")

        template_dict = {key: value for key, value in self.data["node"]["template"].items() if isinstance(value, dict)}
        params = self.params.copy() if self.params else {}

        for edge in self.edges:
            if not hasattr(edge, "target_param"):
                continue
            param_key = edge.target_param

            # If the param_key is in the template_dict and the edge.target_id is the current node
            # We check this to make sure params with the same name but different target_id
            # don't get overwritten
            if param_key in template_dict and edge.target_id == self.id:
                if template_dict[param_key]["list"]:
                    if param_key not in params:
                        params[param_key] = []
                    params[param_key].append(self.graph.get_vertex(edge.source_id))
                elif edge.target_id == self.id:
                    params[param_key] = self.graph.get_vertex(edge.source_id)

        for key, value in template_dict.items():
            if key in params:
                continue
            # Skip _type and any value that has show == False and is not code
            # If we don't want to show code but we want to use it
            if key == "_type" or (not value.get("show") and key != "code"):
                continue
            # If the type is not transformable to a python base class
            # then we need to get the edge that connects to this node
            if value.get("type") == "file":
                # Load the type in value.get('fileTypes') using
                # what is inside value.get('content')
                # value.get('value') is the file name
                if file_path := value.get("file_path"):
                    params[key] = file_path
                else:
                    raise ValueError(f"File path not found for {self.vertex_type}")
            elif value.get("type") in DIRECT_TYPES and params.get(key) is None:
                val = value.get("value")
                if value.get("type") == "code":
                    try:
                        params[key] = ast.literal_eval(val) if val else None
                    except Exception as exc:
                        logger.debug(f"Error parsing code: {exc}")
                        params[key] = val
                elif value.get("type") in ["dict", "NestedDict"]:
                    # When dict comes from the frontend it comes as a
                    # list of dicts, so we need to convert it to a dict
                    # before passing it to the build method
                    if isinstance(val, list):
                        params[key] = {k: v for item in value.get("value", []) for k, v in item.items()}
                    elif isinstance(val, dict):
                        params[key] = val
                elif value.get("type") == "int" and val is not None:
                    try:
                        params[key] = int(val)
                    except ValueError:
                        params[key] = val
                elif value.get("type") == "float" and val is not None:
                    try:
                        params[key] = float(val)
                    except ValueError:
                        params[key] = val
                elif val is not None and val != "":
                    params[key] = val

            if not value.get("required") and params.get(key) is None:
                if value.get("default"):
                    params[key] = value.get("default")
                else:
                    params.pop(key, None)
        # Add _type to params
        self._raw_params = params
        self.params = params

    async def _build(self, user_id=None):
        """
        Initiate the build process.
        """
        logger.debug(f"Building {self.vertex_type}")
        await self._build_each_node_in_params_dict(user_id)
        await self._get_and_instantiate_class(user_id)
        self._validate_built_object()

        self._built = True

    async def _build_each_node_in_params_dict(self, user_id=None):
        """
        Iterates over each node in the params dictionary and builds it.
        """
        for key, value in self.params.copy().items():
            if self._is_node(value):
                if value == self:
                    del self.params[key]
                    continue
                await self._build_node_and_update_params(key, value, user_id)
            elif isinstance(value, list) and self._is_list_of_nodes(value):
                await self._build_list_of_nodes_and_update_params(key, value, user_id)

    def _is_node(self, value):
        """
        Checks if the provided value is an instance of Vertex.
        """
        return isinstance(value, Vertex)

    def _is_list_of_nodes(self, value):
        """
        Checks if the provided value is a list of Vertex instances.
        """
        return all(self._is_node(node) for node in value)

    async def get_result(self, user_id=None, timeout=None) -> Any:
        # Check if the Vertex was built already
        if self._built:
            return self._built_object

        if self.is_task and self.task_id is not None:
            task = self.get_task()

            result = task.get(timeout=timeout)
            if isinstance(result, Coroutine):
                result = await result
            if result is not None:  # If result is ready
                self._update_built_object_and_artifacts(result)
                return self._built_object
            else:
                # Handle the case when the result is not ready (retry, throw exception, etc.)
                pass

        # If there's no task_id, build the vertex locally
        await self.build(user_id=user_id)
        return self._built_object

    async def _build_node_and_update_params(self, key, node, user_id=None):
        """
        Builds a given node and updates the params dictionary accordingly.
        """

        result = await node.get_result(user_id)
        self._handle_func(key, result)
        if isinstance(result, list):
            self._extend_params_list_with_result(key, result)
        self.params[key] = result

    async def _build_list_of_nodes_and_update_params(self, key, nodes: List["Vertex"], user_id=None):
        """
        Iterates over a list of nodes, builds each and updates the params dictionary.
        """
        self.params[key] = []
        for node in nodes:
            built = await node.get_result(user_id)
            if isinstance(built, list):
                if key not in self.params:
                    self.params[key] = []
                self.params[key].extend(built)
            else:
                self.params[key].append(built)

    def _handle_func(self, key, result):
        """
        Handles 'func' key by checking if the result is a function and setting it as coroutine.
        """
        if key == "func":
            if not isinstance(result, types.FunctionType):
                if hasattr(result, "run"):
                    result = result.run  # type: ignore
                elif hasattr(result, "get_function"):
                    result = result.get_function()  # type: ignore
            elif inspect.iscoroutinefunction(result):
                self.params["coroutine"] = result
            else:
                self.params["coroutine"] = sync_to_async(result)

    def _extend_params_list_with_result(self, key, result):
        """
        Extends a list in the params dictionary with the given result if it exists.
        """
        if isinstance(self.params[key], list):
            self.params[key].extend(result)

    async def _get_and_instantiate_class(self, user_id=None):
        """
        Gets the class from a dictionary and instantiates it with the params.
        """
        if self.base_type is None:
            raise ValueError(f"Base type for node {self.vertex_type} not found")
        try:
            result = await loading.instantiate_class(
                node_type=self.vertex_type,
                base_type=self.base_type,
                params=self.params,
                user_id=user_id,
            )
            self._update_built_object_and_artifacts(result)
        except Exception as exc:
            logger.exception(exc)
            raise ValueError(f"Error building node {self.vertex_type}(ID:{self.id}): {str(exc)}") from exc

    def _update_built_object_and_artifacts(self, result):
        """
        Updates the built object and its artifacts.
        """
        if isinstance(result, tuple):
            self._built_object, self.artifacts = result
        else:
            self._built_object = result

    def _validate_built_object(self):
        """
        Checks if the built object is None and raises a ValueError if so.
        """
        if isinstance(self._built_object, UnbuiltObject):
            raise ValueError(f"{self.vertex_type}: {self._built_object_repr()}")
        elif self._built_object is None:
            message = f"{self.vertex_type} returned None."
            if self.base_type == "custom_components":
                message += " Make sure your build method returns a component."

            logger.warning(message)

    async def build(self, force: bool = False, user_id=None, *args, **kwargs) -> Any:
        if not self._built or force:
            await self._build(user_id, *args, **kwargs)

        return self._built_object

    def add_edge(self, edge: "Edge") -> None:
        if edge not in self.edges:
            self.edges.append(edge)

    def __repr__(self) -> str:
        return f"Vertex(id={self.id}, data={self.data})"

    def __eq__(self, __o: object) -> bool:
        try:
            return self.id == __o.id if isinstance(__o, Vertex) else False
        except AttributeError:
            return False

    def __hash__(self) -> int:
        return id(self)

    def _built_object_repr(self):
        # Add a message with an emoji, stars for sucess,
        return "Built sucessfully ✨" if self._built_object is not None else "Failed to build 😵‍💫"



"""Adds Credential table

Revision ID: c1c8e217a069
Revises: 7d2162acc8b2
Create Date: 2023-11-24 10:45:38.465302

"""

from typing import Sequence, Union

import sqlalchemy as sa
import sqlmodel
from alembic import op
from sqlalchemy.engine.reflection import Inspector

# revision identifiers, used by Alembic.
revision: str = "2ac71eb9c3ae"
down_revision: Union[str, None] = "7d2162acc8b2"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    conn = op.get_bind()
    inspector = Inspector.from_engine(conn)  # type: ignore
    tables = inspector.get_table_names()
    try:
        if "credential" not in tables:
            op.create_table(
                "credential",
                sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
                sa.Column("value", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
                sa.Column(
                    "provider", sqlmodel.sql.sqltypes.AutoString(), nullable=True
                ),
                sa.Column("user_id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
                sa.Column("id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
                sa.Column("created_at", sa.DateTime(), nullable=False),
                sa.Column("updated_at", sa.DateTime(), nullable=True),
                sa.PrimaryKeyConstraint("id"),
            )
    except Exception as e:
        print(e)

        pass
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    try:
        op.drop_table("credential")
    except Exception as e:
        print(e)
        pass
    # ### end Alembic commands ###



from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy import func
from sqlalchemy.exc import IntegrityError
from sqlmodel import Session, select
from sqlmodel.sql.expression import SelectOfScalar

from langflow.api.v1.schemas import UsersResponse
from langflow.services.auth.utils import (
    get_current_active_superuser,
    get_current_active_user,
    get_password_hash,
    verify_password,
)
from langflow.services.database.models.user import User, UserCreate, UserRead, UserUpdate
from langflow.services.database.models.user.crud import get_user_by_id, update_user
from langflow.services.deps import get_session, get_settings_service

router = APIRouter(tags=["Users"], prefix="/users")


@router.post("/", response_model=UserRead, status_code=201)
def add_user(
    user: UserCreate,
    session: Session = Depends(get_session),
    settings_service=Depends(get_settings_service),
) -> User:
    """
    Add a new user to the database.
    """
    new_user = User.model_validate(user, from_attributes=True)
    try:
        new_user.password = get_password_hash(user.password)
        new_user.is_active = settings_service.auth_settings.NEW_USER_IS_ACTIVE
        session.add(new_user)
        session.commit()
        session.refresh(new_user)
    except IntegrityError as e:
        session.rollback()
        raise HTTPException(status_code=400, detail="This username is unavailable.") from e

    return new_user


@router.get("/whoami", response_model=UserRead)
def read_current_user(
    current_user: User = Depends(get_current_active_user),
) -> User:
    """
    Retrieve the current user's data.
    """
    return current_user


@router.get("/", response_model=UsersResponse)
def read_all_users(
    skip: int = 0,
    limit: int = 10,
    _: Session = Depends(get_current_active_superuser),
    session: Session = Depends(get_session),
) -> UsersResponse:
    """
    Retrieve a list of users from the database with pagination.
    """
    query: SelectOfScalar = select(User).offset(skip).limit(limit)
    users = session.exec(query).fetchall()

    count_query = select(func.count()).select_from(User)  # type: ignore
    total_count = session.exec(count_query).first()

    return UsersResponse(
        total_count=total_count,  # type: ignore
        users=[UserRead(**user.model_dump()) for user in users],
    )


@router.patch("/{user_id}", response_model=UserRead)
def patch_user(
    user_id: UUID,
    user_update: UserUpdate,
    user: User = Depends(get_current_active_user),
    session: Session = Depends(get_session),
) -> User:
    """
    Update an existing user's data.
    """
    if not user.is_superuser and user.id != user_id:
        raise HTTPException(status_code=403, detail="You don't have the permission to update this user")
    if user_update.password:
        if not user.is_superuser:
            raise HTTPException(status_code=400, detail="You can't change your password here")
        user_update.password = get_password_hash(user_update.password)

    if user_db := get_user_by_id(session, user_id):
        return update_user(user_db, user_update, session)
    else:
        raise HTTPException(status_code=404, detail="User not found")


@router.patch("/{user_id}/reset-password", response_model=UserRead)
def reset_password(
    user_id: UUID,
    user_update: UserUpdate,
    user: User = Depends(get_current_active_user),
    session: Session = Depends(get_session),
) -> User:
    """
    Reset a user's password.
    """
    if user_id != user.id:
        raise HTTPException(status_code=400, detail="You can't change another user's password")

    if not user:
        raise HTTPException(status_code=404, detail="User not found")
    if verify_password(user_update.password, user.password):
        raise HTTPException(status_code=400, detail="You can't use your current password")
    new_password = get_password_hash(user_update.password)
    user.password = new_password
    session.commit()
    session.refresh(user)

    return user


@router.delete("/{user_id}", response_model=dict)
def delete_user(
    user_id: UUID,
    current_user: User = Depends(get_current_active_superuser),
    session: Session = Depends(get_session),
) -> dict:
    """
    Delete a user from the database.
    """
    if current_user.id == user_id:
        raise HTTPException(status_code=400, detail="You can't delete your own user account")
    elif not current_user.is_superuser:
        raise HTTPException(status_code=403, detail="You don't have the permission to delete this user")

    user_db = session.exec(select(User).where(User.id == user_id)).first()
    if not user_db:
        raise HTTPException(status_code=404, detail="User not found")

    session.delete(user_db)
    session.commit()

    return {"detail": "User deleted"}



from typing import Optional

from langchain.text_splitter import Language
from langchain_core.documents import Document

from langflow import CustomComponent


class LanguageRecursiveTextSplitterComponent(CustomComponent):
    display_name: str = "Language Recursive Text Splitter"
    description: str = "Split text into chunks of a specified length based on language."
    documentation: str = "https://docs.langflow.org/components/text-splitters#languagerecursivetextsplitter"

    def build_config(self):
        options = [x.value for x in Language]
        return {
            "documents": {
                "display_name": "Documents",
                "info": "The documents to split.",
            },
            "separator_type": {
                "display_name": "Separator Type",
                "info": "The type of separator to use.",
                "field_type": "str",
                "options": options,
                "value": "Python",
            },
            "separators": {
                "display_name": "Separators",
                "info": "The characters to split on.",
                "is_list": True,
            },
            "chunk_size": {
                "display_name": "Chunk Size",
                "info": "The maximum length of each chunk.",
                "field_type": "int",
                "value": 1000,
            },
            "chunk_overlap": {
                "display_name": "Chunk Overlap",
                "info": "The amount of overlap between chunks.",
                "field_type": "int",
                "value": 200,
            },
            "code": {"show": False},
        }

    def build(
        self,
        documents: list[Document],
        chunk_size: Optional[int] = 1000,
        chunk_overlap: Optional[int] = 200,
        separator_type: str = "Python",
    ) -> list[Document]:
        """
        Split text into chunks of a specified length.

        Args:
            separators (list[str]): The characters to split on.
            chunk_size (int): The maximum length of each chunk.
            chunk_overlap (int): The amount of overlap between chunks.
            length_function (function): The function to use to calculate the length of the text.

        Returns:
            list[str]: The chunks of text.
        """
        from langchain.text_splitter import RecursiveCharacterTextSplitter

        # Make sure chunk_size and chunk_overlap are ints
        if isinstance(chunk_size, str):
            chunk_size = int(chunk_size)
        if isinstance(chunk_overlap, str):
            chunk_overlap = int(chunk_overlap)

        splitter = RecursiveCharacterTextSplitter.from_language(
            language=Language(separator_type),
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
        )

        docs = splitter.split_documents(documents)
        return docs



from typing import List, Optional

from langchain_community.vectorstores.mongodb_atlas import MongoDBAtlasVectorSearch

from langflow import CustomComponent
from langflow.field_typing import Document, Embeddings, NestedDict


class MongoDBAtlasComponent(CustomComponent):
    display_name = "MongoDB Atlas"
    description = "a `MongoDB Atlas Vector Search` vector store from raw documents."

    def build_config(self):
        return {
            "documents": {"display_name": "Documents", "is_list": True},
            "embedding": {"display_name": "Embedding"},
            "collection_name": {"display_name": "Collection Name"},
            "db_name": {"display_name": "Database Name"},
            "index_name": {"display_name": "Index Name"},
            "mongodb_atlas_cluster_uri": {"display_name": "MongoDB Atlas Cluster URI"},
            "search_kwargs": {"display_name": "Search Kwargs", "advanced": True},
        }

    def build(
        self,
        embedding: Embeddings,
        documents: Optional[List[Document]] = None,
        collection_name: str = "",
        db_name: str = "",
        index_name: str = "",
        mongodb_atlas_cluster_uri: str = "",
        search_kwargs: Optional[NestedDict] = None,
    ) -> MongoDBAtlasVectorSearch:
        search_kwargs = search_kwargs or {}
        vector_store = MongoDBAtlasVectorSearch.from_connection_string(
            connection_string=mongodb_atlas_cluster_uri,
            namespace=f"{db_name}.{collection_name}",
            embedding=embedding,
            index_name=index_name,
        )

        if documents is not None:
            if len(documents) == 0:
                raise ValueError("If documents are provided, there must be at least one document.")

            vector_store.add_documents(documents)

        return vector_store



from typing import Callable, Union

from langchain_community.utilities.wolfram_alpha import WolframAlphaAPIWrapper
from langflow import CustomComponent

# Since all the fields in the JSON have show=False, we will only create a basic component
# without any configurable fields.


class WolframAlphaAPIWrapperComponent(CustomComponent):
    display_name = "WolframAlphaAPIWrapper"
    description = "Wrapper for Wolfram Alpha."

    def build_config(self):
        return {"appid": {"display_name": "App ID", "type": "str", "password": True}}

    def build(self, appid: str) -> Union[Callable, WolframAlphaAPIWrapper]:
        return WolframAlphaAPIWrapper(wolfram_alpha_appid=appid)  # type: ignore



from typing import Callable, Optional, Union

from langchain.chains import LLMChain, LLMMathChain

from langflow import CustomComponent
from langflow.field_typing import BaseLanguageModel, BaseMemory, Chain


class LLMMathChainComponent(CustomComponent):
    display_name = "LLMMathChain"
    description = "Chain that interprets a prompt and executes python code to do math."
    documentation = "https://python.langchain.com/docs/modules/chains/additional/llm_math"

    def build_config(self):
        return {
            "llm": {"display_name": "LLM"},
            "llm_chain": {"display_name": "LLM Chain"},
            "memory": {"display_name": "Memory"},
            "input_key": {"display_name": "Input Key"},
            "output_key": {"display_name": "Output Key"},
        }

    def build(
        self,
        llm: BaseLanguageModel,
        llm_chain: LLMChain,
        input_key: str = "question",
        output_key: str = "answer",
        memory: Optional[BaseMemory] = None,
    ) -> Union[LLMMathChain, Callable, Chain]:
        return LLMMathChain(llm=llm, llm_chain=llm_chain, input_key=input_key, output_key=output_key, memory=memory)



from typing import List, Union

from langchain.schema import BaseRetriever
from langchain_community.vectorstores import VectorStore
from langchain_community.vectorstores.supabase import SupabaseVectorStore
from langflow import CustomComponent
from langflow.field_typing import Document, Embeddings, NestedDict
from supabase.client import Client, create_client


class SupabaseComponent(CustomComponent):
    display_name = "Supabase"
    description = "Return VectorStore initialized from texts and embeddings."

    def build_config(self):
        return {
            "documents": {"display_name": "Documents"},
            "embedding": {"display_name": "Embedding"},
            "query_name": {"display_name": "Query Name"},
            "search_kwargs": {"display_name": "Search Kwargs", "advanced": True},
            "supabase_service_key": {"display_name": "Supabase Service Key"},
            "supabase_url": {"display_name": "Supabase URL"},
            "table_name": {"display_name": "Table Name", "advanced": True},
        }

    def build(
        self,
        embedding: Embeddings,
        documents: List[Document],
        query_name: str = "",
        search_kwargs: NestedDict = {},
        supabase_service_key: str = "",
        supabase_url: str = "",
        table_name: str = "",
    ) -> Union[VectorStore, SupabaseVectorStore, BaseRetriever]:
        supabase: Client = create_client(supabase_url, supabase_key=supabase_service_key)
        return SupabaseVectorStore.from_documents(
            documents=documents,
            embedding=embedding,
            query_name=query_name,
            search_kwargs=search_kwargs,
            client=supabase,
            table_name=table_name,
        )



from typing import TYPE_CHECKING
from langflow.services.session.service import SessionService
from langflow.services.factory import ServiceFactory

if TYPE_CHECKING:
    from langflow.services.cache.service import BaseCacheService


class SessionServiceFactory(ServiceFactory):
    def __init__(self):
        super().__init__(SessionService)

    def create(self, cache_service: "BaseCacheService"):
        return SessionService(cache_service)



from typing import List

from langchain import document_loaders
from langchain_core.documents import Document

from langflow import CustomComponent


class UrlLoaderComponent(CustomComponent):
    display_name: str = "Url Loader"
    description: str = "Generic Url Loader Component"

    def build_config(self):
        return {
            "web_path": {
                "display_name": "Url",
                "required": True,
            },
            "loader": {
                "display_name": "Loader",
                "is_list": True,
                "required": True,
                "options": [
                    "AZLyricsLoader",
                    "CollegeConfidentialLoader",
                    "GitbookLoader",
                    "HNLoader",
                    "IFixitLoader",
                    "IMSDbLoader",
                    "WebBaseLoader",
                ],
                "value": "WebBaseLoader",
            },
            "code": {"show": False},
        }

    def build(self, web_path: str, loader: str) -> List[Document]:
        try:
            loader_instance = getattr(document_loaders, loader)(web_path=web_path)
        except Exception as e:
            raise ValueError(f"No loader found for: {web_path}") from e
        docs = loader_instance.load()
        avg_length = sum(len(doc.page_content) for doc in docs if hasattr(doc, "page_content")) / len(docs)
        self.status = f"""{len(docs)} documents)
        \nAvg. Document Length (characters): {int(avg_length)}
        Documents: {docs[:3]}..."""
        return docs



from typing import Optional

from langflow.template.field.base import TemplateField
from langflow.template.frontend_node.base import FrontendNode
from langflow.template.frontend_node.constants import QA_CHAIN_TYPES
from langflow.template.template.base import Template


class ChainFrontendNode(FrontendNode):
    def add_extra_fields(self) -> None:
        if self.template.type_name == "ConversationalRetrievalChain":
            # add memory
            self.template.add_field(
                TemplateField(
                    field_type="BaseChatMemory",
                    required=True,
                    show=True,
                    name="memory",
                    advanced=False,
                )
            )
            # add return_source_documents
            self.template.add_field(
                TemplateField(
                    field_type="bool",
                    required=False,
                    show=True,
                    name="return_source_documents",
                    advanced=False,
                    value=True,
                    display_name="Return source documents",
                )
            )
            self.template.add_field(
                TemplateField(
                    field_type="str",
                    required=True,
                    is_list=True,
                    show=True,
                    multiline=False,
                    options=QA_CHAIN_TYPES,
                    value=QA_CHAIN_TYPES[0],
                    name="chain_type",
                    advanced=False,
                )
            )

    @staticmethod
    def format_field(field: TemplateField, name: Optional[str] = None) -> None:
        FrontendNode.format_field(field, name)
        key = field.name or ""
        if "name" == "RetrievalQA" and key == "memory":
            field.show = False
            field.required = False

        field.advanced = False
        if "key" in key:
            field.password = False
            field.show = False
        if key in ["input_key", "output_key"]:
            field.required = True
            field.show = True
            field.advanced = True

        # We should think of a way to deal with this later
        # if field.field_type == "PromptTemplate":
        #     field.field_type = "str"
        #     field.multiline = True
        #     field.show = True
        #     field.advanced = False
        #     field.value = field.value.template

        # Separated for possible future changes
        if key == "prompt" and field.value is None:
            field.required = True
            field.show = True
            field.advanced = False
        if key == "memory":
            # field.required = False
            field.show = True
            field.advanced = False
        if key == "verbose":
            field.required = False
            field.show = False
            field.advanced = True
        if key == "llm":
            field.required = True
            field.show = True
            field.advanced = False
            field.field_type = "BaseLanguageModel"  # temporary fix
            field.is_list = False

        if key == "return_source_documents":
            field.required = False
            field.show = True
            field.advanced = True
            field.value = True


class SeriesCharacterChainNode(FrontendNode):
    name: str = "SeriesCharacterChain"
    template: Template = Template(
        type_name="SeriesCharacterChain",
        fields=[
            TemplateField(
                field_type="str",
                required=True,
                placeholder="",
                is_list=False,
                show=True,
                advanced=False,
                multiline=False,
                name="character",
            ),
            TemplateField(
                field_type="str",
                required=True,
                placeholder="",
                is_list=False,
                show=True,
                advanced=False,
                multiline=False,
                name="series",
            ),
            TemplateField(
                field_type="BaseLanguageModel",
                required=True,
                placeholder="",
                is_list=False,
                show=True,
                advanced=False,
                multiline=False,
                name="llm",
                display_name="LLM",
            ),
        ],
    )
    description: str = (
        "SeriesCharacterChain is a chain you can use to have a conversation with a character from a series."  # noqa
    )
    base_classes: list[str] = [
        "LLMChain",
        "BaseCustomChain",
        "Chain",
        "ConversationChain",
        "SeriesCharacterChain",
        "Callable",
    ]


class TimeTravelGuideChainNode(FrontendNode):
    name: str = "TimeTravelGuideChain"
    template: Template = Template(
        type_name="TimeTravelGuideChain",
        fields=[
            TemplateField(
                field_type="BaseLanguageModel",
                required=True,
                placeholder="",
                is_list=False,
                show=True,
                advanced=False,
                multiline=False,
                name="llm",
                display_name="LLM",
            ),
            TemplateField(
                field_type="BaseChatMemory",
                required=False,
                show=True,
                name="memory",
                advanced=False,
            ),
        ],
    )
    description: str = "Time travel guide chain."
    base_classes: list[str] = [
        "LLMChain",
        "BaseCustomChain",
        "TimeTravelGuideChain",
        "Chain",
        "ConversationChain",
    ]


class MidJourneyPromptChainNode(FrontendNode):
    name: str = "MidJourneyPromptChain"
    template: Template = Template(
        type_name="MidJourneyPromptChain",
        fields=[
            TemplateField(
                field_type="BaseLanguageModel",
                required=True,
                placeholder="",
                is_list=False,
                show=True,
                advanced=False,
                multiline=False,
                name="llm",
                display_name="LLM",
            ),
            TemplateField(
                field_type="BaseChatMemory",
                required=False,
                show=True,
                name="memory",
                advanced=False,
            ),
        ],
    )
    description: str = "MidJourneyPromptChain is a chain you can use to generate new MidJourney prompts."
    base_classes: list[str] = [
        "LLMChain",
        "BaseCustomChain",
        "Chain",
        "ConversationChain",
        "MidJourneyPromptChain",
    ]


class CombineDocsChainNode(FrontendNode):
    name: str = "CombineDocsChain"
    template: Template = Template(
        type_name="load_qa_chain",
        fields=[
            TemplateField(
                field_type="str",
                required=True,
                is_list=True,
                show=True,
                multiline=False,
                options=QA_CHAIN_TYPES,
                value=QA_CHAIN_TYPES[0],
                name="chain_type",
                advanced=False,
            ),
            TemplateField(
                field_type="BaseLanguageModel",
                required=True,
                show=True,
                name="llm",
                display_name="LLM",
                advanced=False,
            ),
        ],
    )
    description: str = """Load question answering chain."""
    base_classes: list[str] = ["BaseCombineDocumentsChain", "Callable"]

    @staticmethod
    def format_field(field: TemplateField, name: Optional[str] = None) -> None:
        # do nothing and don't return anything
        pass



from typing import List, Optional, Union

import chromadb  # type: ignore
from langchain.embeddings.base import Embeddings
from langchain.schema import BaseRetriever, Document
from langchain_community.vectorstores import VectorStore
from langchain_community.vectorstores.chroma import Chroma
from langflow import CustomComponent


class ChromaComponent(CustomComponent):
    """
    A custom component for implementing a Vector Store using Chroma.
    """

    display_name: str = "Chroma"
    description: str = "Implementation of Vector Store using Chroma"
    documentation = "https://python.langchain.com/docs/integrations/vectorstores/chroma"
    beta: bool = True

    def build_config(self):
        """
        Builds the configuration for the component.

        Returns:
        - dict: A dictionary containing the configuration options for the component.
        """
        return {
            "collection_name": {"display_name": "Collection Name", "value": "langflow"},
            "persist": {"display_name": "Persist"},
            "persist_directory": {"display_name": "Persist Directory"},
            "code": {"advanced": True, "display_name": "Code"},
            "documents": {"display_name": "Documents", "is_list": True},
            "embedding": {"display_name": "Embedding"},
            "chroma_server_cors_allow_origins": {
                "display_name": "Server CORS Allow Origins",
                "advanced": True,
            },
            "chroma_server_host": {"display_name": "Server Host", "advanced": True},
            "chroma_server_port": {"display_name": "Server Port", "advanced": True},
            "chroma_server_grpc_port": {
                "display_name": "Server gRPC Port",
                "advanced": True,
            },
            "chroma_server_ssl_enabled": {
                "display_name": "Server SSL Enabled",
                "advanced": True,
            },
        }

    def build(
        self,
        collection_name: str,
        persist: bool,
        embedding: Embeddings,
        chroma_server_ssl_enabled: bool,
        persist_directory: Optional[str] = None,
        documents: Optional[List[Document]] = None,
        chroma_server_cors_allow_origins: Optional[str] = None,
        chroma_server_host: Optional[str] = None,
        chroma_server_port: Optional[int] = None,
        chroma_server_grpc_port: Optional[int] = None,
    ) -> Union[VectorStore, BaseRetriever]:
        """
        Builds the Vector Store or BaseRetriever object.

        Args:
        - collection_name (str): The name of the collection.
        - persist_directory (Optional[str]): The directory to persist the Vector Store to.
        - chroma_server_ssl_enabled (bool): Whether to enable SSL for the Chroma server.
        - persist (bool): Whether to persist the Vector Store or not.
        - embedding (Optional[Embeddings]): The embeddings to use for the Vector Store.
        - documents (Optional[Document]): The documents to use for the Vector Store.
        - chroma_server_cors_allow_origins (Optional[str]): The CORS allow origins for the Chroma server.
        - chroma_server_host (Optional[str]): The host for the Chroma server.
        - chroma_server_port (Optional[int]): The port for the Chroma server.
        - chroma_server_grpc_port (Optional[int]): The gRPC port for the Chroma server.

        Returns:
        - Union[VectorStore, BaseRetriever]: The Vector Store or BaseRetriever object.
        """

        # Chroma settings
        chroma_settings = None

        if chroma_server_host is not None:
            chroma_settings = chromadb.config.Settings(
                chroma_server_cors_allow_origins=chroma_server_cors_allow_origins or None,
                chroma_server_host=chroma_server_host,
                chroma_server_port=chroma_server_port or None,
                chroma_server_grpc_port=chroma_server_grpc_port or None,
                chroma_server_ssl_enabled=chroma_server_ssl_enabled,
            )

        # If documents, then we need to create a Chroma instance using .from_documents
        if documents is not None and embedding is not None:
            if len(documents) == 0:
                raise ValueError("If documents are provided, there must be at least one document.")
            return Chroma.from_documents(
                documents=documents,  # type: ignore
                persist_directory=persist_directory if persist else None,
                collection_name=collection_name,
                embedding=embedding,
                client_settings=chroma_settings,
            )

        return Chroma(persist_directory=persist_directory, client_settings=chroma_settings)



import datetime
import secrets
import threading
from typing import List, Optional
from uuid import UUID

from sqlmodel import Session, select
from sqlmodel.sql.expression import SelectOfScalar

from langflow.services.database.models.api_key import ApiKey, ApiKeyCreate, ApiKeyRead, UnmaskedApiKeyRead


def get_api_keys(session: Session, user_id: UUID) -> List[ApiKeyRead]:
    query: SelectOfScalar = select(ApiKey).where(ApiKey.user_id == user_id)
    api_keys = session.exec(query).all()
    return [ApiKeyRead.model_validate(api_key) for api_key in api_keys]


def create_api_key(session: Session, api_key_create: ApiKeyCreate, user_id: UUID) -> UnmaskedApiKeyRead:
    # Generate a random API key with 32 bytes of randomness
    generated_api_key = f"sk-{secrets.token_urlsafe(32)}"

    api_key = ApiKey(
        api_key=generated_api_key,
        name=api_key_create.name,
        user_id=user_id,
    )

    session.add(api_key)
    session.commit()
    session.refresh(api_key)
    unmasked = UnmaskedApiKeyRead.model_validate(api_key, from_attributes=True)
    unmasked.api_key = generated_api_key
    return unmasked


def delete_api_key(session: Session, api_key_id: UUID) -> None:
    api_key = session.get(ApiKey, api_key_id)
    if api_key is None:
        raise ValueError("API Key not found")
    session.delete(api_key)
    session.commit()


def check_key(session: Session, api_key: str) -> Optional[ApiKey]:
    """Check if the API key is valid."""
    query: SelectOfScalar = select(ApiKey).where(ApiKey.api_key == api_key)
    api_key_object: Optional[ApiKey] = session.exec(query).first()
    if api_key_object is not None:
        threading.Thread(
            target=update_total_uses,
            args=(
                session,
                api_key_object,
            ),
        ).start()
    return api_key_object


def update_total_uses(session, api_key: ApiKey):
    """Update the total uses and last used at."""
    # This is running in a separate thread to avoid slowing down the request
    # but session is not thread safe so we need to create a new session

    with Session(session.get_bind()) as new_session:
        new_api_key = new_session.get(ApiKey, api_key.id)
        if new_api_key is None:
            raise ValueError("API Key not found")
        new_api_key.total_uses += 1
        new_api_key.last_used_at = datetime.datetime.now(datetime.timezone.utc)
        new_session.add(new_api_key)
        new_session.commit()
    return new_api_key



from langflow import CustomComponent
from langchain.embeddings import VertexAIEmbeddings
from typing import Optional, List


class VertexAIEmbeddingsComponent(CustomComponent):
    display_name = "VertexAIEmbeddings"
    description = "Google Cloud VertexAI embedding models."

    def build_config(self):
        return {
            "credentials": {"display_name": "Credentials", "value": "", "file_types": [".json"], "field_type": "file"},
            "instance": {"display_name": "instance", "advanced": True, "field_type": "dict"},
            "location": {"display_name": "Location", "value": "us-central1", "advanced": True},
            "max_output_tokens": {"display_name": "Max Output Tokens", "value": 128},
            "max_retries": {"display_name": "Max Retries", "value": 6, "advanced": True},
            "model_name": {"display_name": "Model Name", "value": "textembedding-gecko"},
            "n": {"display_name": "N", "value": 1, "advanced": True},
            "project": {"display_name": "Project", "advanced": True},
            "request_parallelism": {"display_name": "Request Parallelism", "value": 5, "advanced": True},
            "stop": {"display_name": "Stop", "advanced": True},
            "streaming": {"display_name": "Streaming", "value": False, "advanced": True},
            "temperature": {"display_name": "Temperature", "value": 0.0},
            "top_k": {"display_name": "Top K", "value": 40, "advanced": True},
            "top_p": {"display_name": "Top P", "value": 0.95, "advanced": True},
        }

    def build(
        self,
        instance: Optional[str] = None,
        credentials: Optional[str] = None,
        location: str = "us-central1",
        max_output_tokens: int = 128,
        max_retries: int = 6,
        model_name: str = "textembedding-gecko",
        n: int = 1,
        project: Optional[str] = None,
        request_parallelism: int = 5,
        stop: Optional[List[str]] = None,
        streaming: bool = False,
        temperature: float = 0.0,
        top_k: int = 40,
        top_p: float = 0.95,
    ) -> VertexAIEmbeddings:
        return VertexAIEmbeddings(
            instance=instance,
            credentials=credentials,
            location=location,
            max_output_tokens=max_output_tokens,
            max_retries=max_retries,
            model_name=model_name,
            n=n,
            project=project,
            request_parallelism=request_parallelism,
            stop=stop,
            streaming=streaming,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
        )



"""Adds tables

Revision ID: 260dbcc8b680
Revises:
Create Date: 2023-08-27 19:49:02.681355

"""

from typing import Sequence, Union

import sqlalchemy as sa
import sqlmodel
from alembic import op
from sqlalchemy.engine.reflection import Inspector

# revision identifiers, used by Alembic.
revision: str = "260dbcc8b680"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###

    conn = op.get_bind()
    inspector = Inspector.from_engine(conn)  # type: ignore
    # List existing tables
    existing_tables = inspector.get_table_names()
    # Drop 'flowstyle' table if it exists
    # and other related indices
    if "flowstyle" in existing_tables:
        op.drop_table("flowstyle")
        if "ix_flowstyle_flow_id" in [
            index["name"] for index in inspector.get_indexes("flowstyle")
        ]:
            op.drop_index(
                "ix_flowstyle_flow_id", table_name="flowstyle", if_exists=True
            )

    existing_indices_flow = []
    existing_fks_flow = []
    if "flow" in existing_tables:
        existing_indices_flow = [
            index["name"] for index in inspector.get_indexes("flow")
        ]
        # Existing foreign keys for the 'flow' table, if it exists
        existing_fks_flow = [
            fk["referred_table"] + "." + fk["referred_columns"][0]
            for fk in inspector.get_foreign_keys("flow")
        ]
        # Now check if the columns user_id exists in the 'flow' table
        # If it does not exist, we need to create the foreign key

    if "user" not in existing_tables:
        op.create_table(
            "user",
            sa.Column("id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
            sa.Column("username", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
            sa.Column("password", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
            sa.Column("is_active", sa.Boolean(), nullable=False),
            sa.Column("is_superuser", sa.Boolean(), nullable=False),
            sa.Column("create_at", sa.DateTime(), nullable=False),
            sa.Column("updated_at", sa.DateTime(), nullable=False),
            sa.Column("last_login_at", sa.DateTime(), nullable=True),
            sa.PrimaryKeyConstraint("id", name="pk_user"),
            sa.UniqueConstraint("id", name="uq_user_id"),
        )
        with op.batch_alter_table("user", schema=None) as batch_op:
            batch_op.create_index(
                batch_op.f("ix_user_username"), ["username"], unique=True
            )

    if "apikey" not in existing_tables:
        op.create_table(
            "apikey",
            sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
            sa.Column("created_at", sa.DateTime(), nullable=False),
            sa.Column("last_used_at", sa.DateTime(), nullable=True),
            sa.Column("total_uses", sa.Integer(), nullable=False, default=0),
            sa.Column("is_active", sa.Boolean(), nullable=False, default=True),
            sa.Column("id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
            sa.Column("api_key", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
            sa.Column("user_id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
            sa.ForeignKeyConstraint(
                ["user_id"], ["user.id"], name="fk_apikey_user_id_user"
            ),
            sa.PrimaryKeyConstraint("id", name="pk_apikey"),
            sa.UniqueConstraint("id", name="uq_apikey_id"),
        )
        with op.batch_alter_table("apikey", schema=None) as batch_op:
            batch_op.create_index(
                batch_op.f("ix_apikey_api_key"), ["api_key"], unique=True
            )
            batch_op.create_index(batch_op.f("ix_apikey_name"), ["name"], unique=False)
            batch_op.create_index(
                batch_op.f("ix_apikey_user_id"), ["user_id"], unique=False
            )
    if "flow" not in existing_tables:
        op.create_table(
            "flow",
            sa.Column("data", sa.JSON(), nullable=True),
            sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
            sa.Column("description", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
            sa.Column("id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
            sa.Column("user_id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
            sa.ForeignKeyConstraint(
                ["user_id"], ["user.id"], name="fk_flow_user_id_user"
            ),
            sa.PrimaryKeyConstraint("id", name="pk_flow"),
            sa.UniqueConstraint("id", name="uq_flow_id"),
        )
    # Conditionally create indices for 'flow' table
    # if _alembic_tmp_flow exists, then we need to drop it first
    # This is to deal with SQLite not being able to ROLLBACK
    # for some unknown reason
    if "_alembic_tmp_flow" in existing_tables:
        op.drop_table("_alembic_tmp_flow")
    with op.batch_alter_table("flow", schema=None) as batch_op:
        flow_columns = [col["name"] for col in inspector.get_columns("flow")]
        if "user_id" not in flow_columns:
            batch_op.add_column(
                sa.Column(
                    "user_id",
                    sqlmodel.sql.sqltypes.GUID(),
                    nullable=True,  # This should be False, but we need to allow NULL values for now
                )
            )
        if "user.id" not in existing_fks_flow:
            batch_op.create_foreign_key("fk_flow_user_id", "user", ["user_id"], ["id"])
        if "ix_flow_description" not in existing_indices_flow:
            batch_op.create_index(
                batch_op.f("ix_flow_description"), ["description"], unique=False
            )
        if "ix_flow_name" not in existing_indices_flow:
            batch_op.create_index(batch_op.f("ix_flow_name"), ["name"], unique=False)
    with op.batch_alter_table("flow", schema=None) as batch_op:
        if "ix_flow_user_id" not in existing_indices_flow:
            batch_op.create_index(
                batch_op.f("ix_flow_user_id"), ["user_id"], unique=False
            )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###

    conn = op.get_bind()
    inspector = Inspector.from_engine(conn)  # type: ignore
    # List existing tables
    existing_tables = inspector.get_table_names()
    if "flow" in existing_tables:
        with op.batch_alter_table("flow", schema=None) as batch_op:
            batch_op.drop_index(batch_op.f("ix_flow_user_id"), if_exists=True)
            batch_op.drop_index(batch_op.f("ix_flow_name"), if_exists=True)
            batch_op.drop_index(batch_op.f("ix_flow_description"), if_exists=True)

        op.drop_table("flow")
    if "apikey" in existing_tables:
        with op.batch_alter_table("apikey", schema=None) as batch_op:
            batch_op.drop_index(batch_op.f("ix_apikey_user_id"), if_exists=True)
            batch_op.drop_index(batch_op.f("ix_apikey_name"), if_exists=True)
            batch_op.drop_index(batch_op.f("ix_apikey_api_key"), if_exists=True)

        op.drop_table("apikey")
    if "user" in existing_tables:
        with op.batch_alter_table("user", schema=None) as batch_op:
            batch_op.drop_index(batch_op.f("ix_user_username"), if_exists=True)

        op.drop_table("user")

    if "flowstyle" in existing_tables:
        op.drop_table("flowstyle")

    if "component" in existing_tables:
        op.drop_table("component")
    # ### end Alembic commands ###



"""User id can be null in Flow

Revision ID: f5ee9749d1a6
Revises: 7843803a87b5
Create Date: 2023-10-18 23:12:27.297016

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "f5ee9749d1a6"
down_revision: Union[str, None] = "7843803a87b5"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    try:
        with op.batch_alter_table("flow", schema=None) as batch_op:
            batch_op.alter_column(
                "user_id", existing_type=sa.CHAR(length=32), nullable=True
            )
    except Exception as e:
        print(e)
        pass

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    try:
        with op.batch_alter_table("flow", schema=None) as batch_op:
            batch_op.alter_column(
                "user_id", existing_type=sa.CHAR(length=32), nullable=False
            )
    except Exception as e:
        print(e)
        pass

    # ### end Alembic commands ###



FORCE_SHOW_FIELDS = [
    "allowed_tools",
    "memory",
    "prefix",
    "examples",
    "temperature",
    "model_name",
    "headers",
    "max_value_length",
    "max_tokens",
    "google_cse_id",
]

DEFAULT_PROMPT = """
I want you to act as a naming consultant for new companies.

Here are some examples of good company names:

- search engine, Google
- social media, Facebook
- video sharing, YouTube

The name should be short, catchy and easy to remember.

What is a good name for a company that makes {product}?
"""

SYSTEM_PROMPT = """
You are a helpful assistant that talks casually about life in general.
You are a good listener and you can talk about anything.
"""

HUMAN_PROMPT = "{input}"

QA_CHAIN_TYPES = ["stuff", "map_reduce", "map_rerank", "refine"]

CTRANSFORMERS_DEFAULT_CONFIG = {
    "top_k": 40,
    "top_p": 0.95,
    "temperature": 0.8,
    "repetition_penalty": 1.1,
    "last_n_tokens": 64,
    "seed": -1,
    "max_new_tokens": 256,
    "stop": None,
    "stream": False,
    "reset": True,
    "batch_size": 8,
    "threads": -1,
    "context_length": -1,
    "gpu_layers": 0,
}

# This variable is used to tell the user
# that it can be changed to use other APIs
# like Prem and LocalAI
OPENAI_API_BASE_INFO = """
The base URL of the OpenAI API. Defaults to https://api.openai.com/v1.

You can change this to use other APIs like JinaChat, LocalAI and Prem.
"""


INPUT_KEY_INFO = """The variable to be used as Chat Input when more than one variable is available."""
OUTPUT_KEY_INFO = """The variable to be used as Chat Output (e.g. answer in a ConversationalRetrievalChain)"""


CLASSES_TO_REMOVE = [
    "RunnableSerializable",
    "Serializable",
    "BaseModel",
    "object",
    "Runnable",
    "Generic",
]



from typing import Optional, Union

from langchain.llms import BaseLLM
from langchain_community.chat_models.openai import ChatOpenAI

from langflow import CustomComponent
from langflow.field_typing import BaseLanguageModel, NestedDict


class ChatOpenAIComponent(CustomComponent):
    display_name = "ChatOpenAI"
    description = "`OpenAI` Chat large language models API."

    def build_config(self):
        return {
            "max_tokens": {
                "display_name": "Max Tokens",
                "field_type": "int",
                "advanced": False,
                "required": False,
            },
            "model_kwargs": {
                "display_name": "Model Kwargs",
                "field_type": "NestedDict",
                "advanced": True,
                "required": False,
            },
            "model_name": {
                "display_name": "Model Name",
                "field_type": "str",
                "advanced": False,
                "required": False,
                "options": [
                    "gpt-4-turbo-preview",
                    "gpt-4-0125-preview",
                    "gpt-4-1106-preview",
                    "gpt-4-vision-preview",
                    "gpt-3.5-turbo-0125",
                    "gpt-3.5-turbo-1106",
                ],
            },
            "openai_api_base": {
                "display_name": "OpenAI API Base",
                "field_type": "str",
                "advanced": False,
                "required": False,
                "info": (
                    "The base URL of the OpenAI API. Defaults to https://api.openai.com/v1.\n\n"
                    "You can change this to use other APIs like JinaChat, LocalAI and Prem."
                ),
            },
            "openai_api_key": {
                "display_name": "OpenAI API Key",
                "field_type": "str",
                "advanced": False,
                "required": False,
                "password": True,
            },
            "temperature": {
                "display_name": "Temperature",
                "field_type": "float",
                "advanced": False,
                "required": False,
                "value": 0.7,
            },
        }

    def build(
        self,
        max_tokens: Optional[int] = 256,
        model_kwargs: NestedDict = {},
        model_name: str = "gpt-4-1106-preview",
        openai_api_base: Optional[str] = None,
        openai_api_key: Optional[str] = None,
        temperature: float = 0.7,
    ) -> Union[BaseLanguageModel, BaseLLM]:
        if not openai_api_base:
            openai_api_base = "https://api.openai.com/v1"
        return ChatOpenAI(
            max_tokens=max_tokens,
            model_kwargs=model_kwargs,
            model=model_name,
            base_url=openai_api_base,
            api_key=openai_api_key,
            temperature=temperature,
        )



import os
from pathlib import Path
import platform

from loguru import logger


def set_secure_permissions(file_path):
    if platform.system() in ["Linux", "Darwin"]:  # Unix/Linux/Mac
        os.chmod(file_path, 0o600)
    elif platform.system() == "Windows":
        import win32api
        import win32con
        import win32security

        user, domain, _ = win32security.LookupAccountName("", win32api.GetUserName())
        sd = win32security.GetFileSecurity(file_path, win32security.DACL_SECURITY_INFORMATION)
        dacl = win32security.ACL()

        # Set the new DACL for the file: read and write access for the owner, no access for everyone else
        dacl.AddAccessAllowedAce(
            win32security.ACL_REVISION,
            win32con.GENERIC_READ | win32con.GENERIC_WRITE,
            user,
        )
        sd.SetSecurityDescriptorDacl(1, dacl, 0)
        win32security.SetFileSecurity(file_path, win32security.DACL_SECURITY_INFORMATION, sd)
    else:
        print("Unsupported OS")


def write_secret_to_file(path: Path, value: str) -> None:
    with path.open("wb") as f:
        f.write(value.encode("utf-8"))
    try:
        set_secure_permissions(path)
    except Exception:
        logger.error("Failed to set secure permissions on secret key")


def read_secret_from_file(path: Path) -> str:
    with path.open("r") as f:
        return f.read()



from typing import Any, Dict, List

from langchain.docstore.document import Document
from langchain.document_loaders.directory import DirectoryLoader
from langflow import CustomComponent


class DirectoryLoaderComponent(CustomComponent):
    display_name = "DirectoryLoader"
    description = "Load from a directory."

    def build_config(self) -> Dict[str, Any]:
        return {
            "glob": {"display_name": "Glob Pattern", "value": "**/*.txt"},
            "load_hidden": {"display_name": "Load Hidden Files", "value": False, "advanced": True},
            "max_concurrency": {"display_name": "Max Concurrency", "value": 10, "advanced": True},
            "metadata": {"display_name": "Metadata", "value": {}},
            "path": {"display_name": "Local Directory"},
            "recursive": {"display_name": "Recursive", "value": True, "advanced": True},
            "silent_errors": {"display_name": "Silent Errors", "value": False, "advanced": True},
            "use_multithreading": {"display_name": "Use Multithreading", "value": True, "advanced": True},
        }

    def build(
        self,
        glob: str,
        path: str,
        max_concurrency: int = 2,
        load_hidden: bool = False,
        recursive: bool = True,
        silent_errors: bool = False,
        use_multithreading: bool = True,
    ) -> List[Document]:
        return DirectoryLoader(
            glob=glob,
            path=path,
            load_hidden=load_hidden,
            max_concurrency=max_concurrency,
            recursive=recursive,
            silent_errors=silent_errors,
            use_multithreading=use_multithreading,
        ).load()



class CustomException(Exception):
    def __init__(self, detail, status_code):
        super().__init__(detail)
        self.status_code = status_code


# Define custom exceptions with status codes
class UnauthorizedError(CustomException):
    def __init__(self, detail="Unauthorized access"):
        super().__init__(detail, 401)


class ForbiddenError(CustomException):
    def __init__(self, detail="Forbidden"):
        super().__init__(detail, 403)


class APIKeyError(CustomException):
    def __init__(self, detail="API key error"):
        super().__init__(detail, 400)  #! Should be 401


class FilterError(CustomException):
    def __init__(self, detail="Filter error"):
        super().__init__(detail, 400)
